import datetime

from pypamguard.chunks.generics import GenericModule, GenericModuleFooter
from .stdmodulefooter import StandardModuleFooter
from .stdmoduleheader import StandardModuleHeader
from .stdchunkinfo import StandardChunkInfo
from pypamguard.core.readers import *
from pypamguard.chunks.standard.stddata import StandardDataMixin
from pypamguard.chunks.annotations.annotationmanager import AnnotationManager

class StandardModule(GenericModule, StandardDataMixin):

    _footer = StandardModuleFooter
    _header = StandardModuleHeader

    def __init__(self, file_header, module_header, filters, *args, **kwargs):
        super().__init__(file_header, module_header, filters, *args, **kwargs)
        self._initialize_stddata()
        self.annotations: AnnotationManager = None

    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        super()._process(br)
        self._process_stddata(br, chunk_info)
    
    def _post(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        super()._post(br)
        if self.flag_bitmap.is_set("HASBINARYANNOTATIONS"):
            self.annotations = AnnotationManager()
            self.annotations.process(br)


from .baseannotation import BaseAnnotation
from pypamguard.core.readers import *

class MatchClsfrAnnotation(BaseAnnotation):

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.data: np.ndarray[np.ndarray[np.float64], np.ndarray[np.float64], np.ndarray[np.float64]] = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        super()._process(br, *args, **kwargs)
        if self.annotation_version == 1:
            # threshold, matchcorr, rejectcorr
            self.data = br.bin_read([DTYPES.FLOAT64, DTYPES.FLOAT64, DTYPES.FLOAT64])
            return
        n_templates = br.bin_read(DTYPES.INT16)        
        self.data = br.bin_read([DTYPES.FLOAT64, DTYPES.FLOAT64, DTYPES.FLOAT64], shape=(n_templates,))
        


"""
A package for registering custom PAMGuard modules, and registering in-built
PAMGuard modules. See `ModuleRegistry`. Effectively
a registration is a mapping from a module name (and possible stream) to a 
class found in `pypamguard.chunks.modules`. When a file header is read, the
registry is used to load the correct module class corresponding to the module
type specified in the header.

`MODULES` dict includes all default mappings. Any
nested dict is a stream-specification rather than just module name.
"""

from pypamguard.chunks.generics import GenericModule
from pypamguard.core.exceptions import ModuleNotFoundException
from pypamguard.logger import logger

from pypamguard.chunks.modules.processing.ais import AISProcessing
__ais_config = AISProcessing

from pypamguard.chunks.modules.detectors.click import ClickDetector
from pypamguard.chunks.modules.detectors.clicktriggerbackground import ClickTriggerBackground
__click_config = {
    "Clicks": ClickDetector,
    "Trigger Background": ClickTriggerBackground,
}

from pypamguard.chunks.modules.processing.clipgenerator import ClipGenerator
__clipgenerator_config = ClipGenerator

from pypamguard.chunks.modules.classifiers.deeplearningclassifier import DLCDetections, DLCModels
__deeplearningclassifier_config = {
    "DL_detection": DLCDetections,
    "DL detection": DLCDetections,
    "DL_Model_Data": DLCModels,
    "DL Model Data": DLCModels,
}

from pypamguard.chunks.modules.processing.dbht import DbHtProcessing
__dbht_config = DbHtProcessing

from pypamguard.chunks.modules.processing.difar import DIFARProcessing
__difar_config = DIFARProcessing

from pypamguard.chunks.modules.processing.noisemonitor import NoiseMonitor
__noisemonitor_config = NoiseMonitor

from pypamguard.chunks.modules.processing.noiseband import NoiseBandMonitor
__noiseband_config = NoiseBandMonitor

from pypamguard.chunks.modules.detectors.gpl import GPLDetector
__gpl_config = GPLDetector

from pypamguard.chunks.modules.detectors.rwedge import RWEdgeDetector
__rwedge_config = RWEdgeDetector

from pypamguard.chunks.modules.detectors.whistleandmoan import WhistleAndMoanDetector
__whistleandmoan_config = WhistleAndMoanDetector

from pypamguard.chunks.modules.processing.longtermspectralaverage import LongTermSpectralAverage
__longtermspectralaverage_config = LongTermSpectralAverage

from pypamguard.chunks.modules.processing.ishmael import IshmaelData, IshmaelDetections
__ishmael_config = {
    "Ishmael Peak Data": IshmaelData, # Not currently output from the detector
    "Ishmael Detections": IshmaelDetections,
}

from pypamguard.chunks.modules.plugins.spermwhaleipi import SpermWhaleIPI
__spermwhaleipi_config = SpermWhaleIPI

from pypamguard.chunks.modules.plugins.geminithreshold import GeminiThresholdDetector
__geminithreshold_config = GeminiThresholdDetector

MODULES = {
    "AIS Processing": __ais_config,
    "Click Detector": __click_config,
    "SoundTrap Click Detector": __click_config,
    "Clip Generator": __clipgenerator_config,
    "Deep Learning Classifier": __deeplearningclassifier_config,
    "DbHt": __dbht_config,
    "DIFAR Processing": __difar_config,
    "LTSA": __longtermspectralaverage_config,
    "Noise Monitor": __noisemonitor_config,
    "Noise Band": __noisemonitor_config,
    "NoiseBand": __noiseband_config,
    "GPL Detector": __gpl_config,
    "RW Edge Detector": __rwedge_config,
    "WhistlesMoans": __whistleandmoan_config,
    "Energy Sum Detector": __ishmael_config,
    "Spectrogram Correlation Detector": __ishmael_config,
    "Matched Filter Detector": __ishmael_config,
    "Ipi module": __spermwhaleipi_config,
    "Gemini Threshold Detector": __geminithreshold_config,
}

class ModuleRegistry:
    """
    A class which stores mappings from module types and streams to module classes
    found in `pypamguard.chunks.modules` (or custom user-defined modules). All
    default modules are in `MODULES`. Usually a module registry is automatically
    created when a `PAMGuard` file is opened. The only time that you might need to
    pass your own `ModuleRegistry` in is when creating custom modules.
    """

    def __init__(self):
        self.modules: dict[str, GenericModule | dict[str, GenericModule]] = {}
        self.__register_preinstalled_modules()

    def register_module(self, module_name: str, module_class: GenericModule | dict):
        """
        Register a new module.
        
        If a module is valid for all streams, then
        ```python
        registry.register_module("Module Type", ModuleClass)
        ```
        
        If a module is stream-specific, then
        ```python
        registry.register_module("Module Type", {"Stream A": ModuleClass, "Stream B": ModuleClass})
        ```
        """
        if module_name in self.modules:
            raise ValueError(f"Module {module_name} is already registered. Deregister module first by calling `deregister_module('{module_name}')`.")
        self.modules[module_name] = module_class
    
    def deregister_module(self, module_name: str) -> int:
        """Deregister a module. Returns the number of modules deregistered (either 0 or 1)"""
        if module_name in self.modules:
            del self.modules[module_name]
            return 1
        return 0
    
    def get_module(self, module_name: str, module_stream) -> GenericModule:
        logger.info(f"{module_name} {module_stream}")
        if module_name in self.modules and type(self.modules[module_name]) == dict:
            if module_stream in self.modules[module_name]: return self.modules[module_name][module_stream]
            raise ModuleNotFoundException(f"Module '{module_name}' is not registered for stream '{module_stream}'.")
        elif module_name in self.modules and issubclass(self.modules[module_name], GenericModule):
            return self.modules[module_name]
        elif module_name in self.modules:
            raise ModuleNotFoundException(f"Module '{module_name}' is not registered correctly.")
        else:
            raise ModuleNotFoundException(f"Module '{module_name}' is not registered.")

    def __register_preinstalled_modules(self):
        for module in MODULES:
            self.register_module(module, MODULES[module])



import io

from pypamguard.chunks.generics import GenericFileHeader
from pypamguard.core.readers import *

class StandardFileHeader(GenericFileHeader):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def _process(self, br: BinaryReader, chunk_info):
        self.file_path = chunk_info.file_path
        self.length = chunk_info.length
        self.identifier = chunk_info.identifier

        self.file_format: int = br.bin_read(DTYPES.INT32)
        self.pamguard: str = br.nstring_read(12)
        self.version: str = br.string_read()
        self.branch: str = br.string_read()
        self.data_date_raw, self.data_date = br.timestamp_read()
        self.analysis_date_raw, self.analysis_date = br.timestamp_read()
        self.start_sample: int = br.bin_read(DTYPES.INT64)
        self.module_type: str = br.string_read()
        self.module_name: str = br.string_read()
        self.stream_name: str = br.string_read()
        self.extra_info_len: int = br.bin_read(DTYPES.INT32)


from .baseannotation import BaseAnnotation
from pypamguard.core.readers import *

class UserFormAnnotation(BaseAnnotation):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form_data: str = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        super()._process(br, *args, **kwargs)
        txt_len = self.annotation_length - len(self.annotation_id) - 2 - 2
        self.form_data = br.nstring_read(txt_len)
        


from contextlib import contextmanager
import io
import time, typing

from pypamguard.core.exceptions import BinaryFileException, WarningException, CriticalException, ChunkLengthMismatch, StructuralException
from pypamguard.chunks.base import BaseChunk
from pypamguard.chunks.standard import StandardBackground, StandardChunkInfo, StandardFileHeader, StandardFileFooter, StandardModuleHeader, StandardModuleFooter
from pypamguard.chunks.generics import GenericBackground, GenericChunkInfo, GenericFileHeader, GenericFileFooter, GenericModuleHeader, GenericModuleFooter, GenericModule
from pypamguard.core.registry import ModuleRegistry
from pypamguard.utils.constants import IdentifierType
from pypamguard.utils.constants import BYTE_ORDERS
from pypamguard.core.filters import Filters, FILTER_POSITION, FilterMismatchException
from pypamguard.logger import logger, Verbosity
from pypamguard.core.serializable import Serializable
from pypamguard.core.readers import *
import os

class FileInfo(Serializable):
    def __init__(self):
        self.__file_header = None
        self.__module_header = None
        self.__module_footer = None
        self.__file_footer = None

    @property
    def file_header(self):
        return self.__file_header

    @file_header.setter
    def file_header(self, file_header: GenericFileHeader):
        self.__file_header = file_header

    @property
    def module_header(self):
        return self.__module_header

    @module_header.setter
    def module_header(self, module_header: GenericModuleHeader):
        self.__module_header = module_header

    @property
    def module_footer(self):
        return self.__module_footer

    @module_footer.setter
    def module_footer(self, module_footer: GenericModuleFooter):
        self.__module_footer = module_footer

    @property
    def file_footer(self):
        return self.__file_footer

    @file_footer.setter
    def file_footer(self, file_footer: GenericFileFooter):
        self.__file_footer = file_footer

class PAMGuardFile(Serializable):
    """
    This class represents a PAMGuard Binary File
    """
    
    def __init__(self, path: str, fp: typing.BinaryIO, order: BYTE_ORDERS = BYTE_ORDERS.BIG_ENDIAN, module_registry: ModuleRegistry = ModuleRegistry(), filters: Filters = Filters(), report: Report = None, clear_fields=None):
        """
        Initialize a PAMGuardFile object. This will set-up the binary reader, but
        not actually read any data yet. See `load()`.

        :param path: The path of the file (this is only used for logging)
        :param fp: The file pointer from which data is read (must be opened with 'rb' mode)
        :param order: Override byte order of the file (optional)
        :param module_registry: Override the module registry (optional)
        :param filters: The filters passed as a `core.filters.Filters` object (optional)
        :param report: A `core.readers.Report` object used for logging (optional)
        """
        if not report: self.report = Report()
        else: self.report = report
        self.__file_info = FileInfo()
        self.__file_info.file_header = StandardFileHeader()
        self.__file_info.file_footer = StandardFileFooter(self.__file_info.file_header)
        self.__path: str = path
        self.__filename = os.path.basename(self.__path)
        self.__fp: io.BufferedReader = fp
        self.__order: BYTE_ORDERS = order
        self.__module_registry: ModuleRegistry = module_registry
        self.__filters: Filters = filters
        self.__module_class: GenericModule.__class__ = None # will be overriden by module registry
        self.__size: int = self.__get_size()

        self.__data: list[GenericModule] = []
        self.__background: list[GenericBackground] = []
        self.__total_time: int = 0
        self.__clear_fields = clear_fields

    def __process_chunk(self, br: BinaryReader, chunk_obj: BaseChunk, chunk_info: GenericChunkInfo, correct_chunk_length = True):
        try:
            if type(chunk_info) in (GenericModule, GenericBackground) and self.__filters.position == FILTER_POSITION.STOP: raise FilterMismatchException()
            logger.debug(f"Processing chunk: {type(chunk_obj)}", br)
            chunk_obj.process(br, chunk_info)
            if not br.at_checkpoint(): raise ChunkLengthMismatch(br, chunk_info, chunk_obj)
        except WarningException as e:
            self.report.add_warning(e)
        except FilterMismatchException as e:
            br.goto_checkpoint()
            return None
        except CriticalException as e:
            raise e
        except Exception as e:
            self.report.add_error(e)
        if correct_chunk_length and not br.at_checkpoint(): br.goto_checkpoint()
        return chunk_obj

    def __get_size(self):
        temp = self.__fp.tell()
        self.__fp.seek(0, io.SEEK_END)
        size = self.__fp.tell()
        self.__fp.seek(temp, io.SEEK_SET)
        return size

    def load(self) -> None:
        """
        Load the PAMGuard Binary File, restarting the `fp` (file pointer) from
        the constructor at the beginning. You can see any warnings or errors using
        the `report` (`pypamguard.core.readers.Report`) attribute.

        Throws a `pypamguard.core.exceptions.BinaryFileException` (or a subclass thereof) if
        something goes wrong.
        """

        start_time = time.time()
        self.__fp.seek(0, io.SEEK_SET)
        data_count = 0
        bg_count = 0

        while True:
            br = BinaryReader(self.__fp, report = self.report)
            if br.tell() == self.__size: break
            
            # each chunk has the same 8-byte 'chunk info' at the start
            chunk_info = StandardChunkInfo(self.__path)
            chunk_info.process(br)
            br.set_checkpoint(chunk_info.length - chunk_info.measured_length)

            logger.debug(f"Reading chunk of type {chunk_info.identifier} and length {chunk_info.length} at offset {br.tell()}", br)

            if chunk_info.identifier == IdentifierType.FILE_HEADER.value:
                self.report.set_context(self.__file_info.file_header.__class__)
                self.__file_info.file_header = self.__process_chunk(br, self.__file_info.file_header, chunk_info, correct_chunk_length=False)
                self.__module_class = self.__module_registry.get_module(self.__file_info.file_header.module_type, self.__file_info.file_header.stream_name)

            elif chunk_info.identifier == IdentifierType.MODULE_HEADER.value:
                self.report.set_context(self.__file_info.module_header.__class__)
                if not self.__file_info.file_header: raise StructuralException(self.__fp, "File header not found before module header")
                self.__file_info.module_header = self.__process_chunk(br, self.__module_class._header(self.__file_info.file_header), chunk_info)

            elif chunk_info.identifier >= 0:
                self.report.set_context(f"{self.__module_class.__class__} [iter {data_count}]")
                if not self.__file_info.module_header: raise StructuralException(self.__fp, "Module header not found before data")
                data = self.__process_chunk(br, self.__module_class(self.__file_info.file_header, self.__file_info.module_header, self.__filters), chunk_info)
                if self.__clear_fields and data:
                    for f in self.__clear_fields:
                        data.__dict__.pop(f, None)
                if data: self.__data.append(data)
                data_count += 1
                
            elif chunk_info.identifier == IdentifierType.MODULE_FOOTER.value:
                self.report.set_context(self.__file_info.module_footer.__class__)
                if not self.__file_info.module_header: raise StructuralException(self.__fp, "Module header not found before module footer")
                self.__file_info.module_footer = self.__process_chunk(br, self.__module_class._footer(self.__file_info.file_header, self.__file_info.module_header), chunk_info)

            elif chunk_info.identifier == IdentifierType.FILE_FOOTER.value:
                self.report.set_context(self.__file_info.file_footer.__class__)
                if not self.__file_info.file_header: raise StructuralException(self.__fp, "File header not found before file footer")
                self.__file_info.file_footer = self.__process_chunk(br, self.__file_info.file_footer, chunk_info)

            elif chunk_info.identifier == IdentifierType.FILE_BACKGROUND.value:
                self.report.set_context(f"{self.__module_class._background.__class__} [iter {bg_count}]")
                if not self.__file_info.module_header: raise StructuralException(self.__fp, "Module header not found before data")
                if self.__module_class._background is None: raise StructuralException(self.__fp, "Module class does not have a background specified")
                background = self.__process_chunk(br, self.__module_class._background(self.__file_info.file_header, self.__file_info.module_header, self.__filters), chunk_info)
                if background: self.__background.append(background)
                bg_count += 1

            elif chunk_info.identifier == IdentifierType.IGNORE.value:
                br.goto_checkpoint()

            else:
                raise StructuralException(self.__fp, f"Unknown chunk identifier: {chunk_info.identifier}")
                
        self.__total_time = time.time() - start_time
        logger.info("File processed in %.2f ms" % (self.__total_time * 1000))

    def to_json(self):
        return {
            "filters": self.filters.to_json() if self.filters else None,
            "file_header": self.__file_info.file_header.to_json() if self.__file_info.file_header else None,
            "module_header": self.__file_info.module_header.to_json() if self.__file_info.module_header else None,
            "module_footer": self.__file_info.module_footer.to_json() if self.__file_info.module_footer else None,
            "file_footer": self.__file_info.file_footer.to_json() if self.__file_info.file_footer else None,
            "data": [chunk.to_json() for chunk in self.__data] if self.__data else [],
            "background": [chunk.to_json() for chunk in self.__background] if self.__background else [],
        }

    def __str__(self):
        ret = f"PAMGuard Binary File (filename={self.__path}, size={self.size} bytes, order={self.__order})\n\n"
        ret += f"{self.__filters}\n"
        ret += f"{self.report}"
        ret += f"File Header\n{self.__file_info.file_header}\n\n"
        ret += f"Module Header\n{self.__file_info.module_header}\n\n"
        ret += f"Module Footer\n{self.__file_info.module_footer}\n\n"
        ret += f"File Footer\n{self.__file_info.file_footer}\n\n"
        ret += f"Data Set: {len(self.__data)} objects\n"
        ret += f"Total time: {self.__total_time:.2f} seconds\n"
        return ret

    def add_file_info(self):
        for data in self.__data:
            data.file_info = self.__file_info
        for background in self.__background:
            background.file_info = self.__file_info

    @property
    def background(self) -> list[GenericBackground]:
        """The background data of the PAMGuard file"""
        return self.__background

    @property
    def data(self) -> list[GenericModule]:
        """The data of the PAMGuard file"""
        return self.__data

    @property
    def filters(self) -> Filters:
        """The filters used when processing the PAMGuard file"""
        return self.__filters

    @property
    def path(self) -> str:
        """The path of the PAMGuard file"""
        return self.__path

    @property
    def file_info(self) -> FileInfo:
        """The file info of the PAMGuard file (headers and footers)"""
        return self.__file_info


from pypamguard.chunks.standard import StandardModule, StandardChunkInfo
from pypamguard.core.readers import *

class DLCModels(StandardModule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.type: np.int8 = None
        self.is_binary: np.uint8 = None
        self.scale: np.float32 = None
        self.n_species: np.int16 = None
        self.predictions: np.ndarray[np.float32] = None
        self.n_class: np.int16 = None

    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        self.type = br.bin_read(DTYPES.INT8)
        self.is_binary = br.bin_read(DTYPES.UINT8) # logical
        self.scale = br.bin_read(DTYPES.FLOAT32)
        self.n_species = br.bin_read(DTYPES.INT16)
        self.predictions = br.bin_read((DTYPES.INT16, lambda x: x / self.scale), shape=(self.n_species,))
        self.n_class = br.bin_read(DTYPES.INT16)
        br.bin_read(DTYPES.INT16, shape=(self.n_class,)) # Not too sure what this is...

class DLCDetections(StandardModule):

    _minimum_version = 2 # As at 9 Jul 2025

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.n_chan: np.int16 = None
        self.n_samps: np.int32 = None
        self.scale: np.float32 = None
        self.wave: np.ndarray[np.float32] = None

    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        super()._process(br, chunk_info)
        self.n_chan = br.bin_read(DTYPES.INT16)
        self.n_samps = br.bin_read(DTYPES.INT32)
        self.scale = br.bin_read((DTYPES.FLOAT32, lambda x: 1/x))
        self.wave = br.bin_read((DTYPES.INT8, lambda x: x / self.scale), shape=(self.n_chan, self.n_samps))



from pypamguard.chunks.base import BaseChunk
from pypamguard.core.readers import *

from .beamformer import BeamFormerAnnotation
from .bearing import BearingAnnotation
from .tm import TMAnnotation
from .tbdl import TBDLAnnotation
from .clickclsfr import ClickClsfrAnnotation
from .matchcls import MatchClsfrAnnotation
from .rwudp import RWUDPAnnotation
from .dl import DLAnnotation
from .userform import UserFormAnnotation

class AnnotationManager(BaseChunk):

    beam_angles: BeamFormerAnnotation
    bearing: BearingAnnotation
    target_motion: TMAnnotation
    toad_angles: TBDLAnnotation
    classification: ClickClsfrAnnotation
    m_classification: MatchClsfrAnnotation
    basic_classification: RWUDPAnnotation
    dl_classification: DLAnnotation
    user_form_data: UserFormAnnotation

    def _process(self, br, *args, **kwargs):
        annotations_length = br.bin_read(DTYPES.INT16)
        n_annotations = br.bin_read(DTYPES.INT16)
        for i in range(n_annotations):
            annotation_length = br.bin_read(DTYPES.INT16) - 2
            annotation_id = br.string_read()
            annotation_version = br.bin_read(DTYPES.INT16)
            kwarg_data = {
                "annotation_length": annotation_length,
                "annotation_id": annotation_id,
                "annotation_version": annotation_version,
            }
            
            if annotation_id == 'Beer':
                self.beam_angles = BeamFormerAnnotation()
                self.beam_angles.process(br, *args, **kwarg_data)
            elif annotation_id == 'Bearing':
                self.bearing = BearingAnnotation()
                self.bearing.process(br, *args, **kwarg_data)
            elif annotation_id == 'TMAN':
                self.target_motion = TMAnnotation()
                self.target_motion.process(br, *args, **kwarg_data)
            elif annotation_id == 'TBDL':
                self.toad_angles = TBDLAnnotation()
                self.toad_angles.process(br, *args, **kwarg_data)
            elif annotation_id == 'ClickClassifier_1':
                self.classification = ClickClsfrAnnotation()
                self.classification.process(br, *args, **kwarg_data)
            elif annotation_id == 'Matched_Clk_Clsfr':
                self.m_classification = MatchClsfrAnnotation()
                self.m_classification.process(br, *args, **kwarg_data)
            elif annotation_id == 'BCLS':
                self.basic_classification = RWUDPAnnotation()
                self.basic_classification.process(br, *args, **kwarg_data)
            elif annotation_id == 'DLRE' or annotation_id == 'Delt':
                self.dl_classification = DLAnnotation()
                self.dl_classification.process(br, *args, **kwarg_data)
            elif annotation_id == 'Uson' or annotation_id == 'USON':
                self.user_form_data = UserFormAnnotation()
                self.user_form_data.process(br, *args, **kwarg_data)
            elif annotation_id == 'USON':
                self.user_form_data = UserFormAnnotation()
                self.user_form_data.process(br, *args, **kwarg_data)
            else:
                raise Exception(f"Unknown annotation type: {annotation_id}")



from abc import ABC, abstractmethod
import io

import numpy as np
from pypamguard.core.serializable import Serializable
from pypamguard.core.readers import BinaryReader

class BaseChunk(Serializable, ABC):

    def __init__(self, *args, **kwargs):
        self._measured_length = None
        self._start_pos = None

    @property
    def measured_length(self):
        return self._measured_length

    def _process(self, br: BinaryReader, *args, **kwargs):
        pass

    def _post(self, br: BinaryReader, *args, **kwargs):
        pass

    def process(self, br: BinaryReader, *args, **kwargs):
        self._start_pos = br.tell()
        self._process(br, *args, **kwargs)
        self._post(br, *args, **kwargs)
        self._measured_length = br.tell() - self._start_pos 

    def get_attrs(self):
        return [attr for attr in self.__dict__ if not attr.startswith('_')]

    def signature(self) -> dict:
        lines = {}
        for attr, value in self.__dict__.items():
            if not attr.startswith('_'):
                lines[attr] = type(value)
        return lines

    def __str__(self):
        lines = []
        for attr, value in self.__dict__.items():
            if not attr.startswith('_') and not value is None:
                lines.append(f"{attr} ({type(value)}): ")
                # Custom code to print the signature of a list
                if isinstance(value, (list, np.ndarray)) and len(value) > 0:
                    shape = []
                    while isinstance(value, (list, np.ndarray)) and len(value) > 0:
                        shape.append(str(len(value)))
                        value = value[0]
                    lines[-1] += f"[{'x'.join(shape)} {value.__class__.__name__}]"
                else:
                    lines[-1] += f"{value}"

        return '\t' + '\n\t'.join(lines)



from abc import ABC, abstractmethod

from pypamguard.chunks.base import BaseChunk

class GenericFileHeader(BaseChunk, ABC):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.length: int = None
        self.identifier: int = None
        self.file_format: int = None
        self.pamguard: str = None
        self.version: str = None
        self.branch: str = None
        self.data_date: int = None
        self.analysis_date: int = None
        self.start_sample: int = None
        self.module_type: str = None
        self.module_name: str = None
        self.stream_name: str = None
        self.extra_info_len: int = None



import time
from contextlib import contextmanager
from pypamguard.logger import logger

@contextmanager
def timer(label):
    logger.debug(f"Started {label}")
    start_time = time.perf_counter()
    yield
    total_time = time.perf_counter() - start_time
    logger.debug(f"Finished {label} in {total_time:.3f} seconds")



import json, datetime, numpy as np
from pypamguard.utils.bitmap import Bitmap

def serialize(value):
    def serialize_list(l):
        if isinstance(l, list):
            return [serialize_list(i) for i in l]
        elif isinstance(l, np.ndarray):
            return serialize_list(l.tolist())
        return l

    if type(value) == np.ndarray: value = serialize_list(value)
    if type(value) == datetime.datetime: value = value.timestamp() * 1000
    if type(value) == Bitmap: value = value.bits
    return value




from .baseannotation import BaseAnnotation
from pypamguard.core.readers import *

class RWUDPAnnotation(BaseAnnotation):

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.label: str = None
        self.method: str = None
        self.score: np.float32 = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        super()._process(br, *args, **kwargs)
        self.label = br.string_read()
        self.method = br.string_read()
        self.score = br.bin_read(DTYPES.FLOAT32)
        


from pypamguard.chunks.standard import StandardModule, StandardModuleFooter, StandardModuleHeader
from pypamguard.chunks.modules.detectors.spectralbackground import SpectralBackground
from pypamguard.core.readers import *

class WhistleAndMoanDetectorHeader(StandardModuleHeader):
    def __init__(self, file_header):
        super().__init__(file_header)

        self.delay_scale: int = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        if self.binary_length != 0:
            self.delay_scale = br.bin_read(DTYPES.INT32)

class WhistleAndMoanDetector(StandardModule):

    _header = WhistleAndMoanDetectorHeader
    _background = SpectralBackground

    def __init__(self, file_header, module_header, filters):
        super().__init__(file_header, module_header, filters)

        self.n_slices: int = None
        self.amplitude: float = None

        self.contour: np.ndarray
        self.contour_width: np.ndarray
        self.slice_numbers: np.ndarray
        self.n_peaks: np.ndarray
        self.peak_data: np.ndarray

    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        # self.n_slices = br.read_numeric(DTYPES.INT16)
        # self.amplitude = br.read_numeric(DTYPES.INT16) / 100

        self.n_slices, self.amplitude = br.bin_read([(DTYPES.INT16), (DTYPES.INT16, lambda x: x / 100)])

        self.slice_numbers = np.ndarray((self.n_slices,), dtype=np.int32)
        self.n_peaks = np.ndarray((self.n_slices,), dtype=np.int8)
        self.peak_data = np.empty(shape=(self.n_slices,), dtype=object)
        self.contour = np.ndarray((self.n_slices), dtype=np.int16)
        self.contour_width = np.ndarray((self.n_slices), dtype=np.int16)

        for i in range(self.n_slices):
            self.slice_numbers[i] = br.bin_read(DTYPES.INT32)
            self.n_peaks[i] = br.bin_read(DTYPES.INT8)
            self.peak_data[i] = br.bin_read(DTYPES.INT16, (self.n_peaks[i], 4))            
            self.contour[i] = self.peak_data[i][0][1]
            self.contour_width[i] = self.peak_data[i][0][2] - self.peak_data[i][0][0] + 1
        
        self.mean_width = self.contour_width.mean()


from pypamguard.chunks.standard import StandardModule, StandardModuleFooter, StandardBackground
from pypamguard.core.readers import *
from pypamguard.logger import logger

class ClickDetectorFooter(StandardModuleFooter):

    def __init__(self, file_header, module_header):
        super().__init__(file_header, module_header)

        self.types_count_length: int = None
        self.types_count: list[int] = None

    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        if self.binary_length > 0:
            self.types_count_length = br.bin_read(DTYPES.INT16)
            if self.types_count_length > 0: self.types_count = br.bin_read(DTYPES.INT32, shape=(self.types_count_length,))
            else: self.types_count = []

class ClickDetectorBackgound(StandardBackground):

    _minimum_version = 2

    def __init__(self, file_header, module_header, filters):
        super().__init__(file_header, module_header, filters)

        self.noise_len: np.int16 = None
        self.background: np.ndarray[np.float32] = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        
        self.noise_len = br.bin_read(DTYPES.INT16)
        self.background = br.bin_read(DTYPES.FLOAT32, shape=(self.noise_len,))

class ClickDetector(StandardModule):

    _minimum_version = 2 # As at 9 Jul 2025
    _footer = ClickDetectorFooter
    _background = ClickDetectorBackgound

    def __init__(self, file_header, module_header, filters):
        super().__init__(file_header, module_header, filters)
        
        self.start_sample: int = None
        self.channel_map: Bitmap = None
        self.trigger_map: Bitmap = None
        self.type: int = None
        self.flags: Bitmap = None
        self.delays: float = None
        self.angles: np.ndarray = np.array([])
        self.angle_errors: np.ndarray = np.array([])
        self.duration: int = None
        self.wave: np.ndarray = None

    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        self.n_chan = len(self.channel_map.get_set_bits())

        if self._module_header.version <= 3:
            self.start_sample, self.channel_map = br.bin_read([DTYPES.INT64, DTYPES.INT32])

        self.trigger_map, self.type = br.bin_read([DTYPES.INT32, DTYPES.INT16])
        self.flags = br.bitmap_read(DTYPES.INT32)
        
        if self._module_header.version <= 3:
            n_delays = br.bin_read(DTYPES.INT16)
            if n_delays: self.delays = br.bin_read(DTYPES.FLOAT32, shape=n_delays)
        
        n_angles = br.bin_read(DTYPES.INT16)
        if n_angles: self.angles = br.bin_read(DTYPES.FLOAT32, shape=n_angles)
        
        n_angle_errors = br.bin_read(DTYPES.INT16)
        if n_angle_errors: self.angle_errors = br.bin_read(DTYPES.FLOAT32, shape=n_angle_errors)
        
        if self._module_header.version <= 3: self.duration = br.bin_read(DTYPES.UINT16)
        else: self.duration = self.sample_duration
        
        max_val = br.bin_read(DTYPES.FLOAT32)
        def normalize_wave(x): return np.round(x * max_val / 127, 4)
        self.wave = br.bin_read((DTYPES.INT8, normalize_wave), shape=(self.n_chan, self.duration))



from pypamguard.chunks.standard import StandardModule
from pypamguard.core.readers import *

class SpermWhaleIPI(StandardModule):

    _minimum_version = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.parent_uid: np.int64 = None
        self.ipi: np.float32 = None
        self.ipi_amplitude: np.float32 = None
        self.sample_rate: np.float32 = None
        self.max_val: np.float32 = None
        self.echo_len: np.int32 = None
        self.echo_data: np.ndarray = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        
        self.parent_uid = br.bin_read(DTYPES.INT64)
        self.ipi = br.bin_read(DTYPES.FLOAT32)
        self.ipi_amplitude = br.bin_read(DTYPES.FLOAT32)
        self.sample_rate = br.bin_read(DTYPES.FLOAT32)
        self.max_val = br.bin_read(DTYPES.FLOAT32)
        self.echo_len = br.bin_read(DTYPES.INT32)
        self.echo_data = br.bin_read((DTYPES.INT16, lambda x: x * self.max_val / 32767), shape=(self.echo_len,))



from pypamguard.chunks.standard import StandardModule, StandardModuleHeader
from pypamguard.core.readers import *

class NoiseMonitorHeader(StandardModuleHeader):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.n_bands: np.int16 = None
        self.stats_types: np.int16 = None
        self.lo_edges: np.ndarray = None
        self.hi_edges: np.ndarray = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        if self.binary_length != 0:
            self.n_bands = br.bin_read(DTYPES.INT16)
            self.stats_types = br.bin_read(DTYPES.INT16)
            self.lo_edges = br.bin_read(DTYPES.FLOAT32, shape=(self.n_bands,))
            self.hi_edges = br.bin_read(DTYPES.FLOAT32, shape=(self.n_bands,))

class NoiseMonitor(StandardModule):

    _minimum_version = 2 # As at 9 Jul 2025
    _header = NoiseMonitorHeader

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.i_chan: np.int16 = None
        self.n_bands: np.int16 = None
        self.n_measures: np.int16 = None
        self.noise: np.ndarray = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        self.i_chan = br.bin_read(DTYPES.INT16)
        self.n_bands = br.bin_read(DTYPES.INT16)
        self.n_measures = br.bin_read(DTYPES.INT16)
        self.noise = br.bin_read((DTYPES.INT16, lambda x: x / 100), shape=(self.n_bands, self.n_measures))



from pypamguard.chunks.generics import GenericModuleHeader
from pypamguard.core.readers import *

class StandardModuleHeader(GenericModuleHeader):
    
    def __init__(self, file_header, *args, **kwargs):
        super().__init__(file_header, *args, **kwargs)
        self.file_path: str = None
        self.length: int = None
        self.identifier: int = None
        self.version: str = None
        self.binary_length: int = None

    def _process(self, br: BinaryReader, chunk_info):
        self.file_path = chunk_info.file_path
        self.length = chunk_info.length
        self.identifier = chunk_info.identifier
        self.version: int = br.bin_read(DTYPES.INT32)
        self.binary_length: int = br.bin_read(DTYPES.INT32)



from .baseannotation import BaseAnnotation
from pypamguard.core.readers import *

class BeamFormerAnnotation(BaseAnnotation):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hydrophones: np.uint32 = None
        self.array_type: np.int16 = None
        self.localisation_content: np.uint32 = None
        self.n_angles: np.int16 = None
        self.angles: np.ndarray[np.float32] = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        super()._process(br, *args, **kwargs)
        self.hydrophones = br.bin_read(DTYPES.UINT32)
        self.array_type = br.bin_read(DTYPES.INT16)
        self.localisation_content = br.bin_read(DTYPES.UINT32)
        self.n_angles = br.bin_read(DTYPES.INT16)
        self.angles = br.bin_read(DTYPES.FLOAT32, shape=(self.n_angles,))



from pypamguard.chunks.standard import StandardModule, StandardBackground
from pypamguard.chunks.modules.detectors.spectralbackground import SpectralBackground
from pypamguard.core.readers import *



class GPLDetector(StandardModule):

    _minimum_version = 0 # As at 9 Jul 2025
    _background = SpectralBackground

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        self.time_res = br.bin_read(DTYPES.FLOAT32)
        self.freq_res = br.bin_read(DTYPES.FLOAT32)
        self.area = br.bin_read(DTYPES.INT16)
        bit_depth = br.bin_read(DTYPES.INT8)
        if bit_depth == 8: p_type = DTYPES.UINT8
        else: p_type = DTYPES.UINT16
        points_x, points_y, self.excess, self.energy = br.bin_read([p_type,p_type, DTYPES.FLOAT32, DTYPES.FLOAT32], shape=(self.area,))
        self.points = np.vstack((points_x, points_y))
        


import io
import numpy as np
import enum, datetime, mmap
from typing import Callable
from pypamguard.utils.constants import DTYPES
from pypamguard.utils.bitmap import Bitmap
from pypamguard.core.exceptions import WarningException, ErrorException, CriticalException, BinaryFileException
from pypamguard.logger import logger
from pypamguard.core.serializable import Serializable
from pypamguard.utils.constants import BYTE_ORDERS
from contextlib import contextmanager
import traceback

class Report(Serializable):
    """
    A class to store warnings and errors during program execution.
    Warnings and errors all extend `pypamguard.core.filters.BinaryFileException`.
    """

    __warnings = list[WarningException]
    __errors: list[ErrorException]
    __errors_tb: list[list[str]]

    @property
    def warnings(self) -> list[WarningException]:
        """The list of warnings"""
        return self.__warnings

    @property
    def errors(self) -> list[ErrorException]:
        """The list of errors"""
        return self.__errors

    @property
    def errors_tb(self) -> list[list[str]]:
        """The list of error tracebacks"""
        return self.__errors_tb

    def __init__(self):
        self.current_context = ""
        self.__warnings = []
        self.__errors = []
        self.__errors_tb = []
    
    def set_context(self, context):
        """Set the current context. Automatically added to
        all warnings and errors when `add_error()` or `add_warning()`
        are called."""
        self.current_context = context

    def add_warning(self, warning: BinaryFileException):
        """Add a warning to the report"""
        warning.add_context(self.current_context)
        self.__warnings.append(warning)
        logger.warning(warning)
    
    def add_error(self, error: Exception):
        """Add an error to the report"""
        if type(error) == BinaryFileException:
            error.add_context(self.current_context)
        self.__errors.append(error)
        logger.error(error)
        tb = traceback.format_stack()
        self.__errors_tb.append(tb)
    
    def __str__(self):
        string = "### REPORT SUMMARY ###\n"
        if len(self.__warnings) != 0:
            string += f" - {len(self.__warnings)} warnings (access warnings via .warnings list).\n"
        if len(self.__errors) != 0:
            string += f" - {len(self.__errors)} errors (access errors via .errors list and tracebacks via .errors_tb parallel list).\n"
        string += "### END OF REPORT ###"
        return string

    def print_error_tb(self, index: int):
        for tb in self.__errors_tb[index]:
            print(tb)
        print(self.__errors[index])

class BinaryReader:
    """
    A class to read data from a binary file. 
    """

    def __init__(self, fp: io.BufferedReader, report: Report, endianess: BYTE_ORDERS = BYTE_ORDERS.BIG_ENDIAN):
        """
        :param fp: The file pointer from which data is read (must be opened with 'rb' mode)
        :param report: A `pypamguard.core.readers.Report` object used for logging in the event of an error
        :param endianess: The endianess of the binary reader
        """
        self.__fp = fp
        self.__endianess = endianess
        self.__report = report
    
    @property
    def endianess(self) -> BYTE_ORDERS:
        """The endianess of the binary reader. All data is read in this endianess."""
        return self.__endianess

    @endianess.setter
    def endianess(self, endianess: BYTE_ORDERS):
        if not type(endianess) == BYTE_ORDERS: raise ValueError("Endianess must be one ENDIANESS enum value.")
        self.__endianess = endianess
    
    @property
    def report(self) -> Report:
        """The report object used for logging in the event of an error in this `BinaryReader`."""
        return self.__report

    def __collate(self, data, dtypes, shape):
        for i, dtype_i in enumerate(dtypes):            
            d = data[f'f{i}'][0] if (len(shape) == 1 and shape[0] == 1) else data[f'f{i}'].reshape(shape)
            yield dtype_i[1](d) if dtype_i[1] is not None else d

    def __read(self, length: int) -> bytes:
        data = self.__fp.read(length)
        return data

    def tell(self) -> int:
        """Return the current file pointer position."""
        return self.__fp.tell()

    def seek(self, offset, whence: int = io.SEEK_SET):
        """Set the file pointer position (should not be used directly)."""
        return self.__fp.seek(offset, whence)

    def set_checkpoint(self, offset: int):
        """Given an offset from the current position of the binary reader, set the
        next checkpoint. Upon calling `goto_checkpoint()`, the binary reader will
        seek to the next checkpoint."""
        self.next_checkpoint = self.tell() + offset

    def goto_checkpoint(self):
        """Seek to the next checkpoint set by `set_checkpoint()`."""
        self.seek(self.next_checkpoint)
    
    def at_checkpoint(self):
        """Return `True` if the binary reader is at the next checkpoint, `False` otherwise."""
        if self.tell() == self.next_checkpoint: return True
        else: return False

    def bin_read(self, dtype: list[tuple[DTYPES, Callable[[np.ndarray], np.ndarray]]], shape: tuple = (1,)) -> int | float | np.ndarray | tuple[np.ndarray]:
        """
        Read data from the file. This function is polymorphic in the sense that it
        can be used for any of the following purposes:

        1. Read in a single value of a given datatype (for example `read_numeric(DTYPES.INT32)`).
        2. Read in an array of values of a given datatype (for example `read_numeric(DTYPES.INT32, (5,))`).
        3. Read in an n-dimensional array of values of a given datatype (for example `read_numeric(DTYPES.INT32, (5, 5))`).
        4. Read in an interleaved array of values of a given datatype (for example `read_numeric([DTYPES.INT32, DTYPES.INT32], (5,))`).

        Read in an array of 5 integers (32-bit).
        Return a single `np.ndarray` of type `np.int32`.
        ```python
        bin_read(DTYPES.INT32, (5,))
        ```

        Read in two 5-length arrays (interleaved int32, int32).
        Return a tuple of two `np.ndarray`s of type `np.int16`
        and `np.int64`.
        ```python
        bin_read([DTYPES.INT16, DTYPES.INT64], (5,))
        ```

        Read in a single float (32-bit) and divide by 100.
        Return a single `np.float32`.
        ```python
        bin_read((DTYPES.FLOAT32, lambda x: x / 100))
        ```

        Read in two 5-length arrays (interleaved float32, int8).
        Divide the int8 array by 100.
        Return a tuple of two `np.ndarray`s of type `np.float32` and `np.float32`.
        (NOTE: the int8 array is returned as a float32 array due to the division by 100.)
        ```python
        bin_read([(DTYPES.FLOAT32), (DTYPES.INT8, lambda x: x/100)], (5,))
        ```

        Read in a single 5x2 array of floats (32-bit).
        Return a 2d `np.ndarray` of type `np.float32`.
        ```python
        bin_read(DTYPES.FLOAT32, (5, 2))
        ```
        """
        if type(shape) != tuple: shape = (shape,)
        dtypes = [(dtype_i, None) if isinstance(dtype_i, DTYPES) else dtype_i for dtype_i in ([dtype] if not isinstance(dtype, list) else dtype)]
        data = np.frombuffer(self.__read(sum(dtype_i[0].value.itemsize for dtype_i in dtypes) * np.prod(shape)   ), dtype=[(f'f{i}', dtype_i[0].value.newbyteorder(self.__endianess.value)) for i, dtype_i in enumerate(dtypes)])
        ret_val = tuple(self.__collate(data, dtypes, shape))
        return ret_val[0] if len(ret_val) == 1 else ret_val

    @classmethod
    def millis_to_timestamp(self, millis: int) -> datetime.datetime:
        """Convert a timestamp in milliseconds to a datetime object."""
        return datetime.datetime.fromtimestamp(millis / 1000, tz=datetime.UTC)

    def timestamp_read(self) -> tuple[int, datetime.datetime]:
        """Read in a timestamp in milliseconds and convert it to a datetime object.
        Return a tuple of (milliseconds, datetime)."""
        with br_report(self):
            millis = self.bin_read(DTYPES.INT64)
            return millis, self.millis_to_timestamp(millis)

    def nstring_read(self, length: int) -> str:
        """Read in a string of length `length` and return it as a string."""
        with br_report(self):
            return self.__read(length).decode("utf-8")

    def string_read(self) -> str:
        """Read in a string. The string length is read at first as a 2-byte integer
        specifying the length."""
        with br_report(self):
            return self.nstring_read(self.bin_read(DTYPES.INT16))
    
    def bitmap_read(self, dtype: DTYPES, labels: list[str] = None) -> Bitmap:
        """Read in a bitmap of type `dtype` and return it as a `pypamguard.utils.bitmap.Bitmap`
        object. Can specify a list of `labels` to be mapped onto each bit in the bitmap. See
        the `Bitmap` class more more information.
        
        Example usage (8-bit bitmap):
        ```python
        bitmap_read(DTYPES.INT8, ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'])
        ```
        """
        with br_report(self):
            return Bitmap(dtype.value.itemsize, labels, int(self.bin_read(dtype)))
    
@contextmanager
def br_report(br: BinaryReader):
    """A context manager that reports warnings and errors to the `BinaryReader` object `br`."""
    try:
        yield
    except WarningException as e:
        br.report.add_warning(e)
    except ErrorException as e:
        br.report.add_error(e)
    except CriticalException as e:
        raise e
    except Exception as e:
        br.report.add_error(ErrorException(br=br, message=str(e)))



from pypamguard.chunks.base.chunk import BaseChunk
import datetime
from pypamguard.utils.bitmap import Bitmap
from pypamguard.core.exceptions import *
from pypamguard.core.readers import *
from pypamguard.chunks.standard import StandardChunkInfo

DATA_FLAG_FIELDS = [
    "TIMEMILLISECONDS",
    "TIMENANOSECONDS",
    "CHANNELMAP",
    "UID",
    "STARTSAMPLE",
    "SAMPLEDURATION",
    "FREQUENCYLIMITS",
    "MILLISDURATION",
    "TIMEDELAYSECONDS",
    "HASBINARYANNOTATIONS",
    "HASSEQUENCEMAP",
    "HASNOISE",
    "HASSIGNAL",
    "HASSIGNALEXCESS"
]

class StandardDataMixin:
    def _initialize_stddata(self, *args, **kwargs):
        self.file_path: str = None
        self.millis: int = None
        self.date: datetime.datetime = None
        self.flags: Bitmap = None
        self.time_ns: int = None
        self.channel_map: Bitmap = None
        self.uid: int = None
        self.start_sample: int = None
        self.sample_duration: int = None
        self.freq_limits: float = None
        self.millis_duration: float = None
        self.time_delays: list[float] = None
        self.sequence_map: float = None
        self.noise: float = None
        self.signal: float = None
        self.signal_excess: float = None
    
    def _process_stddata(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        self.file_path = chunk_info.file_path

        self.millis, self.date = br.timestamp_read()        
        self._filters.filter('daterange', self.date)

        self.flag_bitmap = br.bitmap_read(DTYPES.INT16, DATA_FLAG_FIELDS)
        set_flags = self.flag_bitmap.get_set_bits()
        
        if "TIMENANOSECONDS" in set_flags:
            self.time_ns = br.bin_read(DTYPES.INT64)
        
        if "CHANNELMAP" in set_flags:
            self.channel_map = br.bitmap_read(DTYPES.INT32)
        
        if "UID" in set_flags:
            self.uid = br.bin_read(DTYPES.INT64)
            self._filters.filter('uidrange', self.uid)
            self._filters.filter('uidlist', self.uid)
        
        if "STARTSAMPLE" in set_flags:
            self.start_sample = br.bin_read(DTYPES.INT64)
        
        if "SAMPLEDURATION" in set_flags:
            self.sample_duration = br.bin_read(DTYPES.INT32)
        
        if "FREQUENCYLIMITS" in set_flags:
            self.freq_limits = br.bin_read(DTYPES.FLOAT32, shape=(2,))

        if "MILLISDURATION" in set_flags:
            self.millis_duration = br.bin_read(DTYPES.FLOAT32)
        
        if "TIMEDELAYSECONDS" in set_flags:
            num_time_delays = br.bin_read(DTYPES.INT16)
            self.time_delays = br.bin_read(DTYPES.FLOAT32, shape=(num_time_delays,))

        if "HASSEQUENCEMAP" in set_flags:
            self.sequence_map = br.bin_read(DTYPES.INT32)

        if "HASNOISE" in set_flags:
            self.noise = br.bin_read(DTYPES.FLOAT32)

        if "HASSIGNAL" in set_flags:
            self.signal = br.bin_read(DTYPES.FLOAT32)

        if "HASSIGNALEXCESS" in set_flags:
            self.signal_excess = br.bin_read(DTYPES.FLOAT32)
            
        data_length = br.bin_read(DTYPES.INT32)
        if data_length <= 0: raise WarningException(br, chunk_info, self)



from abc import ABC, abstractmethod
import io

from pypamguard.chunks.base import BaseChunk

class GenericChunkInfo(BaseChunk, ABC):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start: int = None



from pypamguard.chunks.standard import StandardModule, StandardChunkInfo, StandardModuleHeader
from pypamguard.core.readers import *

class ClickTriggerBackgroundHeader(StandardModuleHeader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        if self.binary_length != 0:
            self.channel_map = br.bitmap_read(DTYPES.INT32)
            self.n_chan = len(self.channel_map.get_set_bits())
            self.calibration = br.bin_read(DTYPES.FLOAT32, shape=(self.n_chan,))


class ClickTriggerBackground(StandardModule):

    _header = ClickTriggerBackgroundHeader
    _minimum_version = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.scale: np.float32 = None
        self.raw_levels: np.ndarray[np.float64] = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        self.scale = br.bin_read(DTYPES.FLOAT32)
        self.raw_levels = br.bin_read((DTYPES.INT16, lambda x: x / self.scale), shape=(self._module_header.n_chan,))
        


# bitmap.py

class Bitmap:
    def __init__(self, size: int = 0, fields: list = None, value: int = 0):
        """
        A class that represents a [labelled] bitmap. A bitmap
        is labelled when a list of fields is provided and each
        bit corresponds to a field.
        
        :param size: the number of bytes in the bitmap
        :param fields: a list of fields to label the bitmap from least significant to most significant.
            (defaults to None, in which case the bitmap is unlabelled and each bit corresponds only to an index)
        :param value: the initial value of the bitmap (defaults to 0)
        """

        self.fields2index = {}
        self.index2fields = {}

        self.size = size
        self.bits = value
    
        if fields is not None and fields is not []:
            if type(fields) != list:
                raise ValueError(f"Labels must be of type list (got {type(fields)})")
            if len(fields) > size * 8:
                raise ValueError(f"Labels must be of length at least {size * 8} (got {len(fields)})")
            for index, field in enumerate(fields):
                self.fields2index[field] = index
                self.index2fields[index] = field

        if type(value) != int:
            raise ValueError(f"Value must be of type int (got {type(value)})")
        if size <= 0:
            raise ValueError(f"Size must be greater than 0 (got {size})")

    def __get_index(self, field_or_index):
        if isinstance(field_or_index, str):
            return self.fields2index[field_or_index]
        else:
            return field_or_index

    def set(self, field_or_index):
        """Sets the field or index in the bitmap. """
        index = self.__get_index(field_or_index)
        if index < 0 or index >= (self.size * 8):
            raise ValueError("Index out of range")
        self.bits |= 1 << index
    
    def clear(self, field_or_index):
        """Clears the field or index in the bitmap. """
        index = self.__get_index(field_or_index)
        if index < 0 or index >= (self.size * 8):
            raise ValueError("Index out of range")
        self.bits &= ~(1 << index)
    
    def is_set(self, field_or_index):
        """Returns True if the field or index is set in the bitmap. """
        index = self.__get_index(field_or_index)
        if index < 0 or index >= (self.size * 8):
            raise ValueError("Index out of range")
        return (self.bits & (1 << index)) != 0

    def get_set_bits(self):
        """Returns a list of fields or indices that are set in the bitmap. """
        return [
            self.index2fields.get(index, index)
            for index in range(self.size * 8)
            if (self.bits & (1 << index)) != 0
        ]
    
    def get_binary_representation(self):
        return bin(self.bits)[2:].zfill(self.size * 8)

    def __len__(self):
        return self.size * 8

    def __repr__(self):
        return f"Bitmap(size={self.size}, bits={self.bits}, set_bits={self.get_set_bits()})"



from . import *
from .logger import logger, Verbosity
from pypamguard.core.filters import Filters, DateFilter, WhitelistFilter
import datetime

if __name__ == "__main__":

    d1 = datetime.datetime.fromtimestamp(1499572333281 / 1000, tz = datetime.UTC)
    d2 = datetime.datetime.fromtimestamp(1499572363281 / 1000, tz = datetime.UTC)

    pg_filters = Filters({
        # 'daterange': DateFilter(d1, d2, ordered=True),
        # 'uidlist': WhitelistFilter([2000006, 2000003])
    })

    # click_v4_test1_daterange = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test1.pgdf", json_path="click_v4_test1.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1504477746918 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1504477758412 / 1000, tz = datetime.UTC), ordered=True)}))
    # click_v4_test1 = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test1.pgdf")

    # print(click_v4_test1.data[0])
    # filterednoisemeasurement_v3_test1 = load_pamguard_binary_file("../tests/dataset/processing/filterednoisemeasurement/filterednoisemeasurement_v3_test1.pgdf", json_path="filterednoisemeasurement_v3_test1.json")

    # clicktriggerbackground = load_pamguard_binary_file("../tests/dataset/detectors/clicktriggerbackground/clicktriggerbackground_v0_test1.pgdf", json_path="clicktriggerbackground_v0_test1.json")
    # clicktriggerbackground_daterange = load_pamguard_binary_file("../tests/dataset/detectors/clicktriggerbackground/clicktriggerbackground_v0_test1.pgdf", json_path="clicktriggerbackground_v0_test1_datetime.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1751975573249 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1751975698249 / 1000, tz = datetime.UTC), ordered=True)}))
    # clicktriggerbackground_uidlist = load_pamguard_binary_file("../tests/dataset/detectors/clicktriggerbackground/clicktriggerbackground_v0_test1.pgdf", json_path="clicktriggerbackground_v0_test1_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([2000028, 2000032])}))

    # spermwhaleipi_v1_test1 = load_pamguard_binary_file("../tests/dataset/spermwhaleipi/spermwhaleipi_v1_test1.pgdf", json_path="spermwhaleipi_v1_test1.json", filters=pg_filters)
    # spermwhaleipi_v1_test1_daterange = load_pamguard_binary_file("../tests/dataset/plugins/spermwhaleipi/spermwhaleipi_v1_test1.pgdf", json_path="spermwhaleipi_v1_test1_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1751975639054 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1751975732265 / 1000, tz = datetime.UTC), ordered=True)}))
    # spermwhaleipi_v1_test1_uidlist = load_pamguard_binary_file("../tests/dataset/spermwhaleipi/spermwhaleipi_v1_tes1_uidlist.pgdf", json_path="spermwhaleipi_v1_test1_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([2000002])}))

    # click_v4_test2 = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test2.pgdf", json_path="click_v4_test2.json")
    # click_v4_test2_daterange = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test2.pgdf", json_path="click_v4_test2_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1751975573249 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1751975698249 / 1000, tz = datetime.UTC), ordered=True)}))
    # click_v4_test2_uidlist = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test2.pgdf", json_path="click_v4_test2_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([2000002])}))

    # click_v4_test3 = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test3.pgdf", json_path="click_v4_test3.json")
    # click_v4_test3_daterange = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test3.pgdf", json_path="click_v4_test3_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1751975573249 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1751975698249 / 1000, tz = datetime.UTC), ordered=True)}))
    # click_v4_test3_uidlist = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test3.pgdf", json_path="click_v4_test3_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([2000002])}))

    # noisemonitor_v2_test1 = load_pamguard_binary_file("../tests/dataset/processing/noisemonitor/noisemonitor_v2_test1.pgdf", json_path="noisemonitor_v2_test1.json")
    # noisemonitor_v2_test1_daterange = load_pamguard_binary_file("../tests/dataset/processing/noisemonitor/noisemonitor_v2_test1.pgdf", json_path="noisemonitor_v2_test1_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1499572713281 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1499572713291 / 1000, tz = datetime.UTC), ordered=True)}))
    # noisemonitor_v2_test1_uidlist = load_pamguard_binary_file("../tests/dataset/processing/noisemonitor/noisemonitor_v2_test1.pgdf", json_path="noisemonitor_v2_test1_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([1000005])}))

    # noiseband_v3_test1_daterange = load_pamguard_binary_file("../tests/dataset/processing/noiseband/noiseband_v3_test1.pgdf", json_path="noiseband_v3_test1_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1499572333281 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1533053591103 / 1000, tz = datetime.UTC), ordered=True)}))

    # noisebandnoise_v3_test1 = load_pamguard_binary_file("../tests/dataset/processing/noiseband/noisebandnoise_v3_test1.pgdf", json_path="noisebandnoise_v3_test1.json")
    # noisebandnoise_v3_test1_daterange = load_pamguard_binary_file("../tests/dataset/processing/noiseband/noisebandnoise_v3_test1.pgdf", json_path="noisebandnoise_v3_test1_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1499572713281 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1499572713291 / 1000, tz = datetime.UTC), ordered=True)}))
    # noisebandnoise_v3_test1_uidlist = load_pamguard_binary_file("../tests/dataset/processing/noiseband/noisebandnoise_v3_test1.pgdf", json_path="noisebandnoise_v3_test1_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([1000005])}))

    # noisebandpulses_v3_test1 = load_pamguard_binary_file("../tests/dataset/processing/noiseband/noisebandpulses_v3_test1.pgdf", json_path="noisebandpulses_v3_test1.json")
    # noisebandpulses_v3_test1_daterange = load_pamguard_binary_file("../tests/dataset/processing/noiseband/noisebandpulses_v3_test1.pgdf", json_path="noisebandpulses_v3_test1_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1499572713281 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1499572713291 / 1000, tz = datetime.UTC), ordered=True)}))
    # noisebandpulses_v3_test1_uidlist = load_pamguard_binary_file("../tests/dataset/processing/noiseband/noisebandpulses_v3_test1.pgdf", json_path="noisebandpulses_v3_test1_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([1000005])}))

    # clipgenerator_v3_test1 = load_pamguard_binary_file("../tests/dataset/processing/clipgenerator/clipgenerator_v3_test1.pgdf", json_path="clipgenerator_v3_test1.json")
    # clipgenerator_v3_test1_daterange = load_pamguard_binary_file("../tests/dataset/processing/clipgenerator/clipgenerator_v3_test1.pgdf", json_path="clipgenerator_v3_test1_daterange.json")
    # clipgenerator_v3_test1_uidlist = load_pamguard_binary_file("../tests/dataset/processing/clipgenerator/clipgenerator_v3_test1.pgdf", json_path="clipgenerator_v3_test1_uidlist.json")

    # clipgenerator_v3_test2 = load_pamguard_binary_file("../tests/dataset/processing/clipgenerator/clipgenerator_v3_test2.pgdf", json_path="clipgenerator_v3_test2.json")
    # clipgenerator_v3_test2_daterange = load_pamguard_binary_file("../tests/dataset/processing/clipgenerator/clipgenerator_v3_test2.pgdf", json_path="clipgenerator_v3_test2_daterange.json")
    # clipgenerator_v3_test2_uidlist = load_pamguard_binary_file("../tests/dataset/processing/clipgenerator/clipgenerator_v3_test2.pgdf", json_path="clipgenerator_v3_test2_uidlist.json")

    # dbht_v2_test1 = load_pamguard_binary_file("../tests/dataset/processing/dbht/dbht_v2_test1.pgdf", json_path="dbht_v2_test1.json")
    # dbht_v2_test1_daterange = load_pamguard_binary_file("../tests/dataset/processing/dbht/dbht_v2_test1.pgdf", json_path="dbht_v2_test1_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1751974755492 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1751974792491 / 1000, tz = datetime.UTC), ordered=True)}))
    # dbht_v2_test1_uidlist = load_pamguard_binary_file("../tests/dataset/processing/dbht/dbht_v2_test1.pgdf", json_path="dbht_v2_test1_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([4000161])}))

    # difar_v2_test1 = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test1.pgdf", json_path="difar_v2_test1.json")
    # difar_v2_test1_daterange = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test1.pgdf", json_path="difar_v2_test1_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1533049096511 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1533049175551 / 1000, tz = datetime.UTC), ordered=True)}))
    # difar_v2_test1_uidlist = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test1.pgdf", json_path="difar_v2_test1_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([2])}))

    # difar_v2_test2 = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test2.pgdf", verbosity=Verbosity.ERROR)
    # print(difar_v2_test2.data[19])
    #print(difar_v2_test2.file_header)
    #print(difar_v2_test2.module_header)
    
    # for i, d in enumerate(difar_v2_test2.data):
    #     if d.channel_map == None:
    #         print(i, d)
    # difar_v2_test2 = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test2.pgdf", json_path="difar_v2_test2.json")
    # difar_v2_test2_daterange = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test2.pgdf", json_path="difar_v2_test2_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1533049238271 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1533049697599 / 1000, tz = datetime.UTC), ordered=True)}))
    # difar_v2_test2_uidlist = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test2.pgdf", json_path="difar_v2_test2_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([2,3,4,5,6,7,8,10])}))

    # difar_v2_test3 = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test3.pgdf", json_path="difar_v2_test3.json")
    # difar_v2_test3_daterange = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test3.pgdf", json_path="difar_v2_test3_daterange.json", filters=Filters({'daterange': DateFilter(datetime.datetime.fromtimestamp(1533052916351 / 1000, tz = datetime.UTC), datetime.datetime.fromtimestamp(1533053591103 / 1000, tz = datetime.UTC), ordered=True)}))
    # difar_v2_test3_uidlist = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test3.pgdf", json_path="difar_v2_test3_uidlist.json", filters=Filters({'uidlist': WhitelistFilter([178])}))

    # ishmaeldetections_energysum_v2_test1 = load_pamguard_binary_file("../tests/dataset/processing/ishmael/ishmaeldetections_energysum_v2_test1.pgdf", json_path="ishmaeldetections_energysum_v2_test1.json")
    # ishmaeldetections_energysum_v2_test2 = load_pamguard_binary_file("../tests/dataset/processing/ishmael/ishmaeldetections_energysum_v2_test2.pgdf", json_path="ishmaeldetections_energysum_v2_test2.json")
    # ishmaeldetections_energysum_v2_test3 = load_pamguard_binary_file("../tests/dataset/processing/ishmael/ishmaeldetections_energysum_v2_test3.pgdf", json_path="ishmaeldetections_energysum_v2_test3.json")

    # ishmaeldetections_matchedfilter_v2_test1 = load_pamguard_binary_file("../tests/dataset/processing/ishmael/ishmaeldetections_matchedfilter_v2_test1.pgdf", json_path="ishmaeldetections_matchedfilter_v2_test1.json")
    # ishmaeldetections_matchedfilter_v2_test2 = load_pamguard_binary_file("../tests/dataset/processing/ishmael/ishmaeldetections_matchedfilter_v2_test2.pgdf", json_path="ishmaeldetections_matchedfilter_v2_test2.json")

    # ishmaeldetections_spectrogramcorrelation_v2_test1 = load_pamguard_binary_file("../tests/dataset/processing/ishmael/ishmaeldetections_spectrogramcorrelation_v2_test1.pgdf", json_path="ishmaeldetections_spectrogramcorrelation_v2_test1.json")
    # ishmaeldetections_spectrogramcorrelation_v2_test2 = load_pamguard_binary_file("../tests/dataset/processing/ishmael/ishmaeldetections_spectrogramcorrelation_v2_test2.pgdf", json_path="ishmaeldetections_spectrogramcorrelation_v2_test2.json")

    # ais_test1 = load_pamguard_binary_file("../tests/dataset/ais/AIS_Processing_AIS_Processing_AIS_Processing_20250714_154433.pgdf", json_path="ais_test1.json")

    # gpl_v2_test1 = load_pamguard_binary_file("../tests/dataset/detectors/gpl/gpl_v2_test1.pgdf", json_path="gpl_v2_test1.json")
    # gpl_v2_test1_background = load_pamguard_binary_file("../tests/dataset/detectors/gpl/gpl_v2_test1.pgnf", json_path="gpl_v2_test1_background.json")

    # gpl_v2_test2 = load_pamguard_binary_file("../tests/dataset/detectors/gpl/gpl_v2_test2.pgdf", json_path="gpl_v2_test2.json")
    # gpl_v2_test2_background = load_pamguard_binary_file("../tests/dataset/detectors/gpl/gpl_v2_test2.pgnf", json_path="gpl_v2_test2_background.json")

    # geminithreshold_test1 = load_pamguard_binary_file("../tests/dataset/plugins/geminithreshold/Gemini_Threshold_Detector_Filtered_Threshold_Detector_Sonar_Tracks_20241015_050000.pgdf", json_path="geminithreshold_test1.json")
    # geminithreshold_test1_background = load_pamguard_binary_file("../tests/dataset/plugins/geminithreshold/Gemini_Threshold_Detector_Filtered_Threshold_Detector_Sonar_Tracks_20241015_050000.pgnf", json_path="geminithreshold_test1_background.json")

    # whistleandmoan_v2_test1_background = load_pamguard_binary_file("../tests/dataset/detectors/whistleandmoan/whistleandmoan_v2_test1.pgnf", json_path="whistleandmoan_v2_test1_background.json")

    # click_v4_test1 = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test1.pgdf", json_path="click_v4_test1.json")
    # click_v4_test1_background = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test1.pgnf", json_path="click_v4_test1_background.json")
    
    # click_v4_test2 = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test2.pgdf", json_path="click_v4_test2.json")
    # click_v4_test2_background = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test2.pgnf", json_path="click_v4_test2_background.json")
    
    # click_v4_test3 = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test3.pgdf", json_path="click_v4_test3.json")
    # click_v4_test3_background = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test3.pgnf", json_path="click_v4_test3_background.json")

    # deeplearningclassifier_v2_test1_detections = load_pamguard_binary_file("../tests/dataset/classifiers/deeplearningclassifier/deeplearningclassifier_v2_test1_detections.pgdf", json_path="deeplearningclassifier_v2_test1_detections.json")
    # deeplearningclassifier_v2_test1_models = load_pamguard_binary_file("../tests/dataset/classifiers/deeplearningclassifier/deeplearningclassifier_v2_test1_models.pgdf", json_path="deeplearningclassifier_v2_test1_models.json")
    # deeplearningclassifier_v2_test1_models = load_pamguard_binary_file("../tests/dataset/classifiers/deeplearningclassifier/deeplearningclassifier_v2_test1_models.pgdf")

    # deeplearningclassifier_v2_test1_models = load_pamguard_binary_file("../tests/dataset/classifiers/deeplearningclassifier/deeplearningclassifier_v2_test1_models.pgdx", verbosity=Verbosity.ERROR)
    # print(deeplearningclassifier_v2_test1_models.data[0])

    # print()
    # deeplearningclassifier_v2_test1_models = load_pamguard_binary_file("../tests/dataset/classifiers/deeplearningclassifier/deeplearningclassifier_v2_test1_models.pgdf", verbosity=Verbosity.ERROR)
    # print(deeplearningclassifier_v2_test1_models.file_header)



    # difar_v2_test3_daterange = load_pamguard_binary_file("../tests/dataset/processing/difar/difar_v2_test2.pgdf",verbosity=Verbosity.DEBUG)
    # from .load_pamguard_binary_folder import load_pamguard_binary_folder
    # ods = load_pamguard_binary_folder("../tests/dataset/detectors/gpl", "gpl*.pgdf")
    # print(ods)

    # # cd = load_pamguard_binary_file("../tests/dataset/detectors/click/click_v4_test1.pgdx", json_path="cd.json", verbosity=Verbosity.DEBUG)
    # print(cd.data)
    # print(cd.to_json())
    # arr_1 = []
    # arr_2 = []
    # arr_3 = []

    # path = "/home/sulli/SUMMER2025/pypamguard/Data/porpoise_data/Binary/20130710"
    # logger.set_verbosity(Verbosity.ERROR)
    # d = load_pamguard_binary_folder(path, "*.pgdf")
    
    # for file in d:
    #     arr_2.append(file)
    #     arr_2.append(file)
    #     arr_1.append(d[file].data[0].uid)
    #     arr_1.append(d[file].data[-1].uid)

    # for fname in arr_2:
    #     fname_new = fname
    #     if len(fname) > 80:
    #         fname_new = fname[len(fname)-80:]
    #     arr_3.append(fname_new)
    # print(arr_3, len(arr_3))
    # print([int(x) for x in arr_1], len(arr_1))


    # fnames = ['Click_Detector_Click_Detector_Clicks_20130710_143021.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143021.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143259.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143259.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142706.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142706.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143151.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143151.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142151.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142151.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143044.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143044.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142557.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142557.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142044.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142044.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143406.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143406.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142512.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142512.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143513.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143513.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142405.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142405.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142258.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142258.pgdf']
    # uids = [7000001, 7000199, 10000001, 10002893, 6000001, 6001532, 9000001, 9002078, 1000001, 1001663, 8000001, 8003488, 5000001, 5008314, 1, 21, 11000001, 11005142, 4000001, 4010565, 12000001, 12008681, 3000001, 3002860, 2000001, 2002301]
    #
    # # fnames = ['Click_Detector_Click_Detector_Clicks_20130710_143021.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143259.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142706.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143151.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142151.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143044.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142557.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142044.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143406.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142512.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_143513.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142405.pgdf', 'Click_Detector_Click_Detector_Clicks_20130710_142258.pgdf']
    # # uids = [7000001, 7000199, 10000001, 10002893, 6000001, 6001532, 9000001, 9002078, 1000001, 1001663, 8000001, 8003488, 5000001, 5008314, 1, 21, 11000001, 11005142, 4000001, 4010565, 12000001, 12008681, 3000001, 3002860, 2000001, 2002301]
    #
    # fnames = ['click_v4_test1.pgdf', 'click_v4_test2.pgdf', 'click_v4_test3.pgdf']
    # uids = [5000001, 11000001, 2000001]
    # d = load_pamguard_multi_file("../tests/dataset/detectors/click", fnames, uids)



from __future__ import annotations

import copy
import glob
from pathlib import Path

from pypamguard.chunks.generics import GenericModule, GenericBackground
from pypamguard.load_pamguard_binary_file import load_pamguard_binary_file
from pypamguard.core.filters import FILTER_POSITION, Filters
from pypamguard.core.readers import Report
import os

from pypamguard.logger import logger


def load_pamguard_binary_folder(directory: str | Path, mask: str, clear_fields: list = None, filters: Filters = None, report: Report = None) -> tuple[list[GenericModule], list[GenericBackground], Report]:
    r"""
    A function to load a number of PAMGuard binary files from a directory. Returns a tuple containing an array
    of `pypamguard.chunks.generics.genmodule.GenericModule` (data) objects and a list of `pypamguard.chunks.generics.genbackground.GenericBackground`
     (background) objects. Each of the data and background objects contain an attribute `file_info` containing a pointer
     to the `pypamguard.core.pamguardfile.FileInfo` object pertaining to that data/background chunk.
    Set the verbosity beforehand using `pypamguard.logger.Logger.set_verbosity`.

    ```python
    from pypamguard.logger import logger, Verbosity
    logger.set_verbosity(Verbosity.WARNING) # ignore INFO and DEBUG
    ```

    Example usage:

    ```python
    from pypamguard import load_pamguard_binary_folder
    data, background, report = load_pamguard_binary_folder("path/to/directory/", "*.pgdf")
    ```

    Example usage with an ordered filter:

    ```python
    import datetime
    from pypamguard import load_pamguard_binary_folder
    from pypamguard.core.filters import Filters, DateFilter
    filter_obj = Filters({"daterange": DateFilter(start_date=datetime.datetime(2022, 1, 1, tz = datetime.UTC), end_date=datetime.datetime(2022, 1, 2, tz = datetime.UTC), ordered=True)})
    data, background, report = load_pamguard_binary_folder("path/to/directory/", "*.pgdf", filters=filter_obj)
    ```

    :param directory: The directory containing the PAMGuard binary files
    :param mask: A glob mask to filter the files to load (e.g. '\*.pgdf' to match all files with the .pgdf extension, or '\*\*/\*.pgdf' to match all files with the .pgdf extension in any subdirectory)
    :param clear_fields: A list of fields to clear from the PAMGuardFile object. These fields will be set to None. This might be useful if you want to remove unecessary data from the PAMGuardFile object.
    :param filters: A `core.filters.Filters` object. Can be None.
    :param report: A `core.readers.Report` object. Can be None.
    """

    data = []
    background = []

    if not os.path.exists(directory):
        raise FileNotFoundError(f"Directory {directory} does not exist.")
    if not report:
        report = Report()

    files = glob.glob(pathname=mask, root_dir=directory, recursive=True)
    for idx, file in enumerate(files):
        filter_copy = copy.deepcopy(filters)
        res = load_pamguard_binary_file(
            os.path.join(directory, file),
            filters=filter_copy,
            report=report,
            clear_fields=clear_fields,
            )
        res.add_file_info()
        data.extend(res.data)
        background.extend(res.background)
        logger.info(f"Processed files: {idx + 1}")
        if res.filters.position == FILTER_POSITION.STOP: break

    return data, background, report



import os
from pathlib import Path
from pypamguard.core.pamguardfile import PAMGuardFile
from pypamguard.utils.constants import BYTE_ORDERS, DEFAULT_BUFFER_SIZE
from pypamguard.core.filters import Filters, DateFilter
from pypamguard.core.readers import Report
from .logger import logger, Verbosity, logger_config
import io, json
from pypamguard.utils.timer import timer

def load_pamguard_binary_file(filename: str | Path, order: BYTE_ORDERS = BYTE_ORDERS.BIG_ENDIAN, buffering: int | None = DEFAULT_BUFFER_SIZE, filters: Filters = None, json_path: str = None, report: Report = None, clear_fields = None) -> PAMGuardFile:
    """
    Read a binary PAMGuard data file into a PAMFile object

    ```python
    import pypamguard
    file = pypamguard.load_binary_data_file('path/to/data/file.pgdf')
    print(file.file_info.file_header)
    print(file.file_info.module_header)
    print(file.file_info.file_footer)
    print(file.file_info.module_footer)
    print(file.data[0])
    ```

    :param filename: absolute or relative path to the .pgdt file to read
    :param order: endianness of data
    :param buffering: number of bytes to buffer
    :param filters: filters to apply to data
    :param json_path: write json to a specific path
    :param report: report object to use for logging
    :param clear_fields: list of fields to remove from data objects in the PAMGuardFile (for memory)
    :return: PAMGuardFile
    """
    if not filters: filters = Filters()

    with timer("loading PAMGuard binary file"):
        if not os.path.exists(filename):
            raise FileNotFoundError(f"File {filename} does not exist")
        with open(filename, "rb", buffering=buffering) as f:
            f = PAMGuardFile(path=filename, fp=f, order=order, filters=filters, report=report, clear_fields=clear_fields)
            f.load()
    if json_path:
        with open(json_path, 'w') as output:
            json_data = json.dumps(f.to_json(), indent=0, separators=(",", ": "))
            with timer(f"writing output JSON to {output.name}"):
                output.write(json_data)
    return f




from pypamguard.chunks.standard import StandardModule, StandardChunkInfo
from pypamguard.core.readers import *

class NoiseBandMonitor(StandardModule):
    
    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        super()._process(br, chunk_info)
        self.rms = br.bin_read((DTYPES.INT16, lambda x: x / 100))
        self.zero_peak = br.bin_read((DTYPES.INT16, lambda x: x / 100))
        self.peak_peak = br.bin_read((DTYPES.INT16, lambda x: x / 100))
        self.sel = br.bin_read((DTYPES.INT16, lambda x: x / 100))
        self.sel_secs = br.sel_secs = br.bin_read(DTYPES.INT16)
        


from abc import ABC, abstractmethod

from pypamguard.chunks.base import BaseChunk

from .genfileheader import GenericFileHeader
from .genmoduleheader import GenericModuleHeader

class GenericModuleFooter(BaseChunk, ABC):
    
    def __init__(self, file_header: GenericFileHeader, module_header: GenericModuleHeader, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._file_header = file_header
        self._module_header = module_header

        self.length: int = None
        self.identifier: int = None



from pypamguard.chunks.standard import StandardBackground
from pypamguard.core.readers import *

class SpectralBackground(StandardBackground):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.first_bin: np.int32 = None
        self.noise_len: np.int32 = None
        self.background: np.ndarray[np.float32] = None

    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        self.first_bin = br.bin_read(DTYPES.INT32)
        self.noise_len = br.bin_read(DTYPES.INT32)
        self.background = br.bin_read(DTYPES.FLOAT32, shape=(self.noise_len,))


from pypamguard.chunks.standard import StandardModule
from pypamguard.core.readers import *

class DbHtProcessing(StandardModule):

    _minimum_version = 2 # As at 9 Jul 2025

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rms: np.float32 = None
        self.zero_peak: np.float32 = None
        self.peak_peak: np.float32 = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        self.rms = br.bin_read((DTYPES.INT16, lambda x: x / 100))
        self.zero_peak = br.bin_read((DTYPES.INT16, lambda x: x / 100))
        self.peak_peak = br.bin_read((DTYPES.INT16, lambda x: x / 100))


from pypamguard.chunks.standard import StandardModule, StandardChunkInfo
from pypamguard.core.readers import *
from pypamguard.logger import logger

class IshmaelDetections(StandardModule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.peak_height: np.float64 = None
        self.time_sample: np.float64 = None

    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        super()._process(br, chunk_info)
        # Ishmael detections contain simply standard module data
        # NOTE: missing 20 bytes of data. Not accounted for in Matlab code.
        self.peak_height = br.bin_read(DTYPES.FLOAT64)
        self.time_sample = br.bin_read(DTYPES.FLOAT64)


class IshmaelData(StandardModule):

    _minimum_version = 2

    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        super()._process(br, chunk_info)
        n_det = br.bin_read(DTYPES.INT32)
        n_det2 = br.bin_read(DTYPES.INT32)
        self.data = br.bin_read(DTYPES.FLOAT64, shape=(n_det, n_det2))


import io
from pypamguard.chunks.standard import StandardModule, StandardBackground, StandardChunkInfo
from pypamguard.chunks.base import BaseChunk
from pypamguard.core.readers import *
import zlib, math


class GeminiThresholdDetectorBackgroundHeader(BaseChunk):
    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        self.m_id_char = br.bin_read(DTYPES.UINT8)
        self.m_version = br.bin_read(DTYPES.UINT8)
        self.m_length = br.bin_read(DTYPES.UINT32)
        self.m_timestamp = br.bin_read(DTYPES.FLOAT64)
        self.m_data_type = br.bin_read(DTYPES.UINT8)
        self.tm_device_id = br.bin_read(DTYPES.UINT16)
        self.m_node_id = br.bin_read(DTYPES.UINT16)
        self.spare = br.bin_read(DTYPES.UINT16)

class GeminiThresholdDetectorBackgroundGLFRecord(BaseChunk):
    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        self.id1 = br.bin_read(DTYPES.UINT16)
        self.efef = hex(br.bin_read(DTYPES.UINT16))
        self.image_version = br.bin_read(DTYPES.UINT16)
        self.start_range = br.bin_read(DTYPES.UINT32)
        self.end_range = br.bin_read(DTYPES.UINT32)
        self.range_compression = br.bin_read(DTYPES.UINT16)
        self.start_bearing = br.bin_read(DTYPES.UINT32)
        self.end_bearing = br.bin_read(DTYPES.UINT32)
        if self.image_version == 3: skip = br.bin_read(DTYPES.UINT16)
        n_bearing = self.end_bearing - self.start_bearing
        n_range = self.end_range - self.start_range
        packed_size = br.bin_read(DTYPES.UINT32)
        zipped = br.bin_read(DTYPES.UINT8, shape=(packed_size,))
        full_len = n_range * n_bearing
        unzipped_bytes = bytearray(zlib.decompressobj().decompress(zipped))
        if len(unzipped_bytes) < full_len:
            n_range = math.floor(len(unzipped_bytes)/n_bearing)
            self.end_range = n_range
            unzipped_bytes = unzipped_bytes[:n_range * n_bearing]
        # Read the unzipped data into a BinaryReader by emulating a buffered reader
        # Note conversion to uint16 to allow for in-place arithmetic and the copy()
        # as np.ndarray from BinaryReader by default is immutable.
        unzipped_reader = BinaryReader(io.BytesIO(unzipped_bytes), br.report, endianess=BYTE_ORDERS.LITTLE_ENDIAN)
        self.image_data = np.uint16(unzipped_reader.bin_read(DTYPES.UINT8, shape=(n_range, n_bearing)).copy())
        is_neg = np.where(self.image_data < 0)[0]
        self.image_data.put(is_neg, self.image_data.take(is_neg) + 256)
        self.bearing_table = br.bin_read(DTYPES.FLOAT64, n_bearing)
        self.state_flags = br.bitmap_read(DTYPES.UINT32)
        self.modulation_frequency = br.bin_read(DTYPES.UINT32)
        self.beam_form_aperture = br.bin_read(DTYPES.FLOAT32)
        self.tx_time = br.bin_read(DTYPES.FLOAT64)
        self.ping_flags = br.bin_read(DTYPES.UINT16)
        self.sos_at_xd = br.bin_read(DTYPES.FLOAT32)
        self.percent_gain = br.bin_read(DTYPES.UINT16)
        self.chirp = br.bin_read(DTYPES.UINT8)
        self.sonar_type = br.bin_read(DTYPES.UINT8)
        self.platform = br.bin_read(DTYPES.UINT8)
        self.one_spare = br.bin_read(DTYPES.UINT8)
        self.dede = hex(br.bin_read(DTYPES.UINT16))
        self.max_range = self.end_range * self.sos_at_xd/2 /self.modulation_frequency
        t_offs = 723181 # 1 Jan 1980
        secs_per_day = 3600*24;
        self.tx_date = (self.tx_time / secs_per_day) + t_offs   

class GeminiThresholdDetectorBackground(StandardBackground):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        br.endianess = BYTE_ORDERS.LITTLE_ENDIAN

        self.head = GeminiThresholdDetectorBackgroundHeader()
        self.head._process(br, chunk_info)

        self.glf = GeminiThresholdDetectorBackgroundGLFRecord()
        self.glf._process(br, chunk_info)


class GeminiThresholdDetector(StandardModule):

    _background = GeminiThresholdDetectorBackground

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.n_points: np.int32 = None
        self.n_sonar: np.int8 = None
        self.sonar_ids: np.ndarray[np.int16] = None
        self.straight_length: np.float32 = None
        self.wobbly_length: np.float32 = None
        self.mean_occupancy: np.float32 = None
        self.time_millis: np.ndarray[np.int64] = None
        self.sonar_id: np.ndarray[np.int16] = None
        self.min_bearing: np.ndarray[np.float32] = None
        self.max_bearing: np.ndarray[np.float32] = None
        self.peak_bearing: np.ndarray[np.float32] = None
        self.min_range: np.ndarray[np.float32] = None
        self.max_range: np.ndarray[np.float32] = None
        self.peak_range: np.ndarray[np.float32] = None
        self.obj_size: np.ndarray[np.float32] = None
        self.occupancy: np.ndarray[np.float32] = None
        self.ave_value: np.ndarray[np.int16] = None
        self.tot_value: np.ndarray[np.int32] = None
        self.max_value: np.ndarray[np.int16] = None
        self.dates: np.ndarray[datetime.datetime] = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        self.n_points = br.bin_read(DTYPES.INT32)
        self.n_sonar = br.bin_read(DTYPES.INT8)
        self.sonar_ids = br.bin_read(DTYPES.INT16, shape=(self.n_sonar,))
        self.straight_length, self.wobbly_length, self.mean_occupancy = br.bin_read([DTYPES.FLOAT32, DTYPES.FLOAT32, DTYPES.FLOAT32])
        (
            self.time_millis,
            self.sonar_id,
            self.min_bearing,
            self.max_bearing,
            self.peak_bearing,
            self.min_range,
            self.max_range,
            self.peak_range,
            self.obj_size,
            self.occupancy,
            self.ave_value,
            self.tot_value,
            self.max_value,
        ) = br.bin_read([
            DTYPES.INT64,
            DTYPES.INT16,
            DTYPES.FLOAT32,
            DTYPES.FLOAT32,
            DTYPES.FLOAT32,
            DTYPES.FLOAT32,
            DTYPES.FLOAT32,
            DTYPES.FLOAT32,
            DTYPES.FLOAT32,
            DTYPES.FLOAT32,
            DTYPES.INT16,
            DTYPES.INT32,
            DTYPES.INT16,
        ], shape=(self.n_points,))
        self.dates = [BinaryReader.millis_to_timestamp(x) for x in self.time_millis]



from pypamguard.chunks.base import BaseChunk
from pypamguard.core.readers import *

class BaseAnnotation(BaseChunk):

    def __init__(self, *args, **kwargs):
        self.annotation_id = None
        self.annotation_length = None
        self.annotation_version = None

    def process(self, br, *args, **kwargs):
        """Require kwargs to contain annotation_id, annotation_length and annotation_version."""
        super().process(br, *args, **kwargs)

    def _process(self, br, *args, **kwargs):
        if not ('annotation_id' in kwargs and 'annotation_length' in kwargs and 'annotation_version' in kwargs):
            raise ValueError('kwargs must contain annotation_id, annotation_length and annotation_version.')
        self.annotation_id = kwargs['annotation_id']
        self.annotation_length = kwargs['annotation_length']
        self.annotation_version = kwargs['annotation_version']



from abc import ABC, abstractmethod
import numpy as np
import datetime
from pypamguard.utils.bitmap import Bitmap

class Serializable(ABC):
    
    def serialize(self, value):

        if issubclass(type(value), Serializable):
            return value.to_json()

        if isinstance(value, (list, set, np.ndarray)):
            if len(value) == 0: return value
            return [self.serialize(i) for i in value]
        if isinstance(value, np.floating): # remove floating point precision errors#
            return round(float(value), np.finfo(value.dtype).precision)
        if type(value) == datetime.datetime: return value.timestamp()
        if type(value) == Bitmap: return value.bits
        if isinstance(value, (int, float, str, bool, type(None))): return value
        elif isinstance(value, np.generic): return value.item()
        return str(value)
    

    def to_json(self):
        return {**{attr: self.serialize(value) for attr, value in self.__dict__.items() if not attr.startswith('_')}, "__name__": self.__class__.__name__}

    @classmethod
    def from_json(cls, json):
        raise NotImplementedError



from pypamguard.chunks.standard import StandardModule
from pypamguard.core.readers import DTYPES, BinaryReader
import numpy as np

class RWEdgeDetector(StandardModule):

    def __init__(self, file_header, module_header, filters):
        super().__init__(file_header, module_header, filters)

        self.type: int = None
        self.signal: float = None
        self.noise: float = None
        self.n_slices: int = None
        self.slice_nums: np.ndarray = None
        self.lo_freqs: np.ndarray = None
        self.peak_freqs: np.ndarray = None
        self.hi_freqs: np.ndarray = None
        self.peak_amp: np.ndarray = None

    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)

        self.type = br.bin_read(DTYPES.INT16)
        self.signal = br.bin_read(DTYPES.FLOAT32)
        self.noise = br.bin_read(DTYPES.FLOAT32)
        self.n_slices = br.bin_read(DTYPES.INT16)

        (self.slice_nums, self.lo_freqs, self.peak_freqs, self.hi_freqs, self.peak_amp) = br.bin_read([DTYPES.INT16, DTYPES.INT16, DTYPES.INT16, DTYPES.INT16, DTYPES.FLOAT32], (self.n_slices,))



from pypamguard.chunks.base import BaseChunk
from pypamguard.core.readers import *

class Location(BaseChunk):
    def __init__(self):
        self.latitude = None
        self.longitude = None
        self.height = None
        self.error = None
    
    def _process(self, br, *args, **kwargs):
        self.latitude = br.bin_read(DTYPES.FLOAT64)
        self.longitude = br.bin_read(DTYPES.FLOAT64)
        self.height = br.bin_read(DTYPES.FLOAT32)
        self.error = br.string_read()




from .stdchunkinfo import StandardChunkInfo
from .stdmodule import StandardModule
from .stdfileheader import StandardFileHeader
from .stdfilefooter import StandardFileFooter
from .stdmoduleheader import StandardModuleHeader
from .stdmodulefooter import StandardModuleFooter
from .stdbackground import StandardBackground



import io
# from pypamguard.core.readers import BinaryReader

class MultiFileException(Exception):

    def __init__(self, fname, *args, **kwargs):
        super().__init__(*args)
        self.fname = fname

    def __str__(self):
        return f"Error reading {self.fname}: {super().__str__()}"


class BinaryFileException(Exception):

    def __init__(self, br):
        super().__init__()
        self.br = br
        self.position = br.tell()
        self.context = ""

    def add_context(self, context):
        self.context = context  

    def __str__(self):
        return f"{self.__class__.__name__}: {self.context} at position {self.position} bytes"

class WarningException(BinaryFileException):
    
    def __init__(self, br):
        super().__init__(br)

    def __str__(self):
        return f"{super().__str__()}"

class ErrorException(BinaryFileException):
    def __init__(self, br, message = ""):
        super().__init__(br)
        self.message = message
    
    def __str__(self):
        return f"{super().__str__()}, reading {self.__class__.__name__}, {self.message}"

class ChunkLengthMismatch(WarningException):
    def __init__(self, br, chunk_info, chunk_obj):
        super().__init__(br)
        self.chunk_info = chunk_info
        self.chunk_obj = chunk_obj
    
    def __str__(self):
        return super().__str__() + f", expected chunk length {self.chunk_info.length if self.chunk_info else 'unknown'} bytes, got {self.chunk_obj._measured_length + self.chunk_info._measured_length if self.chunk_obj else 'unknown'} bytes"

class CriticalException(BinaryFileException):
    pass

class ModuleNotFoundException(Exception):
    pass

class FileCorruptedException(CriticalException):
    pass

class StructuralException(FileCorruptedException):
    
    def __init__(self, file: io.BufferedReader, message=None):
        super().__init__(file)
        self.message = message
    
    def __str__(self):
        return f"{super().__str__()}, {self.message}"

class NoFileHeaderException(StructuralException):
    pass

class NoModuleHeaderException(StructuralException):
    pass

class NoModuleFooterException(StructuralException):
    pass

class NoFileFooterException(StructuralException):
    pass




from .baseannotation import BaseAnnotation
from pypamguard.core.readers import *

class ClickClsfrAnnotation(BaseAnnotation):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n_classification: np.int16 = None
        self.classify_set: np.ndarray[np.int16] = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        super()._process(br, *args, **kwargs)
        self.n_classifications = br.bin_read(DTYPES.INT16)
        self.classify_set = br.bin_read(DTYPES.INT16, shape=(self.n_classifications,))


import datetime
import enum
from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import Callable
from pypamguard.core.serializable import Serializable

class FilterMismatchException(Exception):
    """
    Raised when a filter is applied to a value and is not met. It is automatically
    thrown by the `Filters` class.
    """
    pass

class FILTER_POSITION(enum.Enum):
    SKIP = 0 # ignore this object but keep reading
    KEEP = 1 # include this object
    STOP = 2 # ignore this, and all following objects

class BaseFilter(Serializable, ABC):
    """Base class for all filters."""

    @abstractmethod
    def check(self, value) -> FILTER_POSITION:
        """
        Apply the filter to the value and return either `KEEP`, `SKIP` or `STOP`
        from the `FILTER_POSITION` enum. 
        """
        raise NotImplementedError

    @classmethod
    def from_json(self, json):
        name = json.pop("__name__")
        if name == "WhitelistFilter":
            return WhitelistFilter.from_json(json)
        elif name == "RangeFilter":
            return RangeFilter.from_json(json)
        elif name == "DateFilter":
            return DateFilter.from_json(json)
        else:
            raise NotImplementedError(f"Implement the .from_json() method for the {name} filter to deserialize from JSON.")

class WhitelistFilter(BaseFilter):
    """A filter that checks if a value is in a set of whitelisted values."""

    def __init__(self, whitelist: set | list):
        """
        Pass in an unordered list or a set of values to include in the whitelist.
        @param whitelist: A list or set of values to include in the whitelist.
        """
        self.whitelist = set(whitelist)

    def check(self, value):
        if value in self.whitelist:
            return FILTER_POSITION.KEEP
        return FILTER_POSITION.SKIP

    def __str__(self):
        return f"WhitelistFilter({len(self.whitelist)} elements)"

    @classmethod
    def from_json(cls, json):
        self = cls(whitelist=set(json["whitelist"]))
        return self

class RangeFilter(BaseFilter):
    def __init__(self, start, end, comparator=None, validation_func=None, ordered=False, ignore_none=False):
        """
        Process a value and return KEEP either when the value is between the start and end values.
        Return SKIP or STOP when the value is outside the range, depending on the ordered flag.

        @param start: The start value of the range.
        @param end: The end value of the range.
        @param comparator: A function that takes two values and returns a boolean (defaults to <=).
        @param validation_func: A function that takes a value and returns a boolean (defaults to None).
        @param ordered: A boolean that indicates if the range is ordered (defaults to False). An ordered filter
                        will offer improved efficiency as once it gets outside of range it will throw a `STOP`.
        @param ignore_none: A boolean that indicates if None values should be ignored (defaults to False). If
                        set to True it will return `KEEP` if the value is None.
        """
        self.start = start
        self.end = end
        self.comparator = comparator or self.default_comparator
        self.validation_func = validation_func
        self.ordered = ordered
        self.ignore_none = ignore_none

    def default_comparator(self, a, b):
        return a < b

    def _validate(self, value):
        if self.validation_func:
            return self.validation_func(value)
        return True

    def check(self, value):
        # if end is 3 and value is 2 then self.comparator(end, value) -> 3 < 2
        if value is None:
            return FILTER_POSITION.KEEP if self.ignore_none else FILTER_POSITION.SKIP
        if not self._validate(value):
            return FILTER_POSITION.SKIP
        if self.comparator(value, self.start):
            return FILTER_POSITION.SKIP
        if self.comparator(self.end, value):
            return FILTER_POSITION.STOP if self.ordered else FILTER_POSITION.SKIP
        return FILTER_POSITION.KEEP
    
    def __str__(self):
        return f"RangeFilter(start={self.start}, end={self.end}, ordered={self.ordered}, ignore_none={self.ignore_none})"

    @classmethod
    def from_json(cls, json):
        return cls(json["start"], json["end"], json["ordered"], json["ignore_none"])

class DateFilter(RangeFilter):
    def __init__(self, start_date, end_date, ordered=False, ignore_timezone=False, ignore_none=False):
        """
        Process a datetime object and return KEEP either when the date
        is between the start and end date or when the date is None. Return
        SKIP when the date is before the start date and STOP when the date
        is after the end date.
        """
        self.ignore_timezone = ignore_timezone
        if not ignore_timezone and start_date.tzinfo is None:
            raise ValueError("Start date must be timezone aware")
        if not ignore_timezone and end_date.tzinfo is None:
            raise ValueError("End date must be timezone aware") 

        super().__init__(start_date, end_date, ordered=ordered, ignore_none=ignore_none)

    def check(self, value: datetime.datetime):
        if value and not self.ignore_timezone and value.tzinfo is None:
            raise ValueError("Value must be timezone aware")
        return super().check(value)

    def __str__(self):
        return f"DateFilter(start={self.start}, end={self.end}, ordered={self.ordered}, ignore_timezone={self.ignore_timezone}, ignore_none={self.ignore_none})"

    @classmethod
    def from_json(cls, json):
        return cls(datetime.datetime.fromtimestamp(json["start"], tz=datetime.UTC), datetime.datetime.fromtimestamp(json["end"], tz=datetime.UTC), json["ordered"], json["ignore_timezone"], json["ignore_none"])

class Filters:
    """
    A class to manage zero or more filters to be applied to a `PAMGuardFile` object.
    Apply a filter to a particular value by calling `filter()` method. Then check the
    `position` (type `FILTER_POSITION`) attribute to check the position of the filter.

    To see all available filters, see attribute `INSTALLED_FILTERS`. You can also add
    your own custom filters using the method `add_custom_filter()` and passing a 
    custom function. 
    """

    INSTALLED_FILTERS = {
        'uidlist': WhitelistFilter,
        'uidrange': RangeFilter,
        'daterange': DateFilter,
    }


    def __init__(self, filters: dict[str, BaseFilter] = None):
        """
        A class to manage zero or more filters to apply to a `PAMGuardFile` object.

        @param filters: A dictionary of filters to apply to a `PAMGuardFile` object.
                        The dictionary key is hte name of the filter (which is important
                        for in-built filters such as `uidlist`) and the value is the
                        filter object (extends `BaseFilter`). Defaults to `None`.
        """

        if not filters: filters = {}
        for key, value in filters.items():
            self.__validate(key, value)
        self.__filters: dict[str, BaseFilter] = filters
        self.position: FILTER_POSITION = None
        self.custom_filters = []
    
    def add_custom_filter(self, func: Callable[[any], bool]):
        """
        Add a custom filter function. The function should take a single argument
        (extends `BaseChunk`) and return a boolean (whether the filter succeeds).
        Custom filters are usually called just after an entire chunk has been read
        and determines if that chunk should be saved or not."""
        self.custom_filters.append(func)
    
    def call_custom_filters(self, chunk_obj):
        for func in self.custom_filters:
            if func(chunk_obj):
                return FILTER_POSITION.KEEP
            else:
                raise FilterMismatchException()

    def add(self, key: str, value: BaseFilter):
        """
        Add a filter. `key` is the name of the filter (which is important
        for in-built filters such as `uidlist`) and the value is the
        filter object (extends `BaseFilter`).
        """
        self.__validate(key, value)
        self.__filters[key] = value
    
    def to_json(self):
        return {
            key: value.to_json() for key, value in self.__filters.items()
        }

    def __validate(self, key, value):
        if key in self.INSTALLED_FILTERS and not isinstance(value, self.INSTALLED_FILTERS[key]):
            raise ValueError(f"Filter {key} must be of type {self.INSTALLED_FILTERS[key]}")
        if key not in self.INSTALLED_FILTERS and not isinstance(value, BaseFilter):
            raise ValueError(f"Custom filter {key} must be of type FilterBinaryFile")

    def filter(self, key: str, value: any) -> None:
        """
        Pass a particular `value` into a filter of type `key`.
        Raises `FilterMismatchException` if the filter is not met

        @param key: The name of the filter to apply.
        @param value: The value to apply the filter to.
        """
        if key in self.__filters:
            self.position = self.__filters[key].check(value)
        else:
            self.position = FILTER_POSITION.KEEP
        
        if not self.position == FILTER_POSITION.KEEP:
            raise FilterMismatchException()

    def __str__(self):
        if len(self.__filters) == 0:
            return "No filters"
        ret = f"Filters ({len(self.__filters)}):\n"
        for filter in self.__filters:
            ret += f"\t{filter}: {self.__filters[filter]}\n"
        return ret
    
    def __len__(self):
        return len(self.__filters)

    @property
    def filters(self):
        return self.__filters


from enum import Enum
import numpy as np

# Arguments that can be used when reading
TIME_RANGE = "timerange"
UID_RANGE = "uidrange"
UID_LIST = "uidlist"
FILTER = "filter"
CHANNEL = "channel"

DEFAULT_BUFFER_SIZE = 1024

class BYTE_ORDERS(Enum):
    LITTLE_ENDIAN = "<"
    BIG_ENDIAN = ">"


class IdentifierType(Enum):
    FILE_HEADER = -1
    FILE_FOOTER = -2
    MODULE_HEADER = -3
    MODULE_FOOTER = -4
    IGNORE = -5
    FILE_BACKGROUND = -6

class DTYPES(Enum):
    INT8 = np.dtype(np.int8)
    UINT8 = np.dtype(np.uint8)
    INT16 = np.dtype(np.int16)
    UINT16 = np.dtype(np.uint16)
    INT32 = np.dtype(np.int32)
    UINT32 = np.dtype(np.uint32)
    INT64 = np.dtype(np.int64)
    UINT64 = np.dtype(np.uint64)
    FLOAT32 = np.dtype(np.float32)
    FLOAT64 = np.dtype(np.float64)



from .genchunkinfo import GenericChunkInfo
from .genfileheader import GenericFileHeader
from .genmoduleheader import GenericModuleHeader
from .genmodulefooter import GenericModuleFooter
from .genfilefooter import GenericFileFooter
from .genmodule import GenericModule
from .genbackground import GenericBackground



import io

from . import StandardChunkInfo
from pypamguard.chunks.generics import GenericFileHeader, GenericFileFooter
from pypamguard.core.readers import *

class StandardFileFooter(GenericFileFooter):
    
    def __init__(self, file_header: GenericFileHeader):
        super().__init__(file_header)
        self.file_footer: str = None
        self.n_objects: int = None
        self.data_date_raw: np.int64 = None
        self.data_date: datetime.datetime = None
        self.analysis_date_raw: np.int64 = None
        self.analysis_date: datetime.datetime = None
        self.end_sample: np.int64 = None
        self.lowest_uid: np.int64 = None
        self.highest_uid: np.int64 = None
        self.file_length: np.int64 = None
        self.end_reason: np.int32 = None

    def _process(self, br: BinaryReader, chunk_info: StandardChunkInfo):
        self.file_path = chunk_info.file_path
        self.length = chunk_info.length
        self.identifier = chunk_info.identifier

        self.n_objects = br.bin_read(DTYPES.INT32)
        self.data_date_raw, self.data_date = br.timestamp_read()
        self.analysis_date_raw, self.analysis_date = br.timestamp_read()
        self.end_sample = br.bin_read(DTYPES.INT64)
        if self._file_header.file_format >= 3:
            self.lowest_uid = br.bin_read(DTYPES.INT64)
            self.highest_uid = br.bin_read(DTYPES.INT64)
        self.file_length = br.bin_read(DTYPES.INT64)
        self.end_reason = br.bin_read(DTYPES.INT32)


"""
.. include:: ./documentation.md
"""
from .load_pamguard_binary_file import load_pamguard_binary_file
from .load_pamguard_binary_folder import load_pamguard_binary_folder
from .load_pamguard_multi_file import load_pamguard_multi_file


from pypamguard.chunks.base import BaseChunk
from pypamguard.core.readers import *

class ModelData(BaseChunk):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.predictions: np.ndarray[np.floating] = None
        self.class_id: np.ndarray[np.int16] = None
        self.is_binary: bool = None
        self.type: np.int18 = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        model_type = br.bin_read(DTYPES.INT8)
        is_binary = bool(br.bin_read(DTYPES.INT8) != 0)
        scale = br.bin_read(DTYPES.FLOAT32)
        n_species = br.bin_read(DTYPES.INT16)
        data = br.bin_read((DTYPES.INT16, lambda x: x/scale), shape=(n_species,))

        n_class = br.bin_read(DTYPES.INT16)
        class_names = br.bin_read(DTYPES.INT16, shape=(n_class,))

        if model_type == 0 or model_type == 1: # generic deep learning annotation
            self.predictions = data
            self.class_id = class_names
            self.is_binary = is_binary
            self.type = model_type
        elif model_type == 2: # dummy result
            self.predictions = np.ndarray(0)
            self.type = 'dummy'


from .baseannotation import BaseAnnotation
from pypamguard.chunks.annotations.submodules import Location
from pypamguard.core.readers import *

class TMAnnotation(BaseAnnotation):

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.model: str = None
        self.n_locations: np.int16 = None
        self.hydrophones: np.uint32 = None
        self.loc: np.ndarray[Location] = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        super()._process(br, *args, **kwargs)
        self.model = br.string_read()
        self.n_locations = br.bin_read(DTYPES.INT16)
        self.hydrophones = br.bin_read(DTYPES.UINT32)
        self.loc = np.ndarray(self.n_locations, type=self.Location)
        for i in range(self.n_locations):
            loc = Location()
            loc.process(br)
            self.loc[i] = loc



from pypamguard.chunks.standard import StandardModule
from pypamguard.core.readers import *


class ClipGenerator(StandardModule):

    _minimum_version = 3 # As at 10 Jul 2025

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.trigger_millis: np.int64 = None
        self.file_name: str = None
        self.trigger_name: str = None
        self.trigger_uid: np.int64 = None
        self.n_chan: np.int16 = None
        self.n_samps: np.int32 = None
        self.scale: np.float32 = None
        self.wave: np.ndarray[np.floating] = None
    
    def _process(self, br, chunk_info):
        
        super()._process(br, chunk_info)

        self.trigger_millis = br.bin_read(DTYPES.INT64)
        self.file_name = br.string_read()
        self.trigger_name = br.string_read()
        self.trigger_uid = br.bin_read(DTYPES.INT64)
        if chunk_info.identifier == 2:
            self.n_chan = br.bin_read(DTYPES.INT16)
            self.n_samps = br.bin_read(DTYPES.INT32)
            self.scale = br.bin_read((DTYPES.FLOAT32, lambda x: 1/x))
            self.wave = br.bin_read((DTYPES.INT8, lambda x: np.int16(x) * self.scale), shape=(self.n_chan, self.n_samps))



import os, glob
from pathlib import Path
from pypamguard.core.filters import Filters, WhitelistFilter
from .load_pamguard_binary_file import load_pamguard_binary_file
from .logger import logger, Verbosity
from pypamguard.chunks.generics import GenericModule
from pypamguard.core.readers import Report
from pypamguard.core.exceptions import CriticalException, MultiFileException

_last_root = None
_last_mask = None
_master_list = []
_master_dict = {}
_MAX_NAME_LEN = 80

def find_binary_file(root, mask, file):
    global _last_mask, _last_root, _master_list, _master_dict
    if (not _last_root or not _last_mask) or (_last_root != root or _last_mask != mask):
        _master_list = glob.glob(pathname=mask, root_dir=root, recursive=True)
        _master_dict = {}
        for reldir in _master_list:
            path = os.path.join(root, reldir)
            fname = os.path.basename(path)
            short_name = fname[len(fname)-_MAX_NAME_LEN:] if len(fname) > _MAX_NAME_LEN else fname
            if short_name not in _master_dict:
                _master_dict[short_name] = path
        _last_root = root
        _last_mask = mask
    if file in _master_dict:
        return _master_dict[file]
    else:
        return None

def load_pamguard_multi_file(data_dir: str | Path, file_names: list[str], item_uids: list[int]) -> tuple[list[GenericModule], Report]:
    """
    A function to load a number of PAMGuard data chunks at once from various binary files, filtering by UID.
    Will return a tuple containing a list of `pypamguard.chunks.generics.GenericModule` objects (event data)
    and a `core.readers.Report` object (with errors/warnings).

    For example, the following code will expect three files, `file1.pgdf`, `file2.pgdf` and `file3.pgdf`
    in the directory `./data` with the respective UIDs.
    ```python
    file_names=["file1.pgdf", "file1.pgdf", "file2.pgdf", "file3.pgdf", "file3.pgdf"]
    item_uids=[7000001, 7000199, 10000001, 10002893, 6000001]
    event_data, report = load_pamguard_multi_file("./data", file_names, item_uids)
    ```

    - A `FileNotFoundError` is raised if `data_dir` does not exist.
    - A `ValueError` is raised if `file_names` and `item_uids` are not the same length.
    - A `FileNotFoundError` is added to the report for each file that is not found.
    - A `pypamguard.core.exceptions.MultiFileException` is added to the report for each file
        that requires one or more UIDs that aren't found.
    - If any warnings/errors occur when reading a file, they are added to the report.
    """
    if not os.path.exists(data_dir):
        raise FileNotFoundError(f"Data directory {data_dir} does not exist.")
    if len(file_names) != len(item_uids):
        raise ValueError("file_names and item_uids must be the same length.")

    file_name_dict = {}
    report = Report()

    event_data = []
    logger.set_verbosity(verbosity=Verbosity.ERROR)
        
    # Each file name has one or more UIDs. Better represented by dict.
    for file_name, uid in zip(file_names, item_uids):
        if file_name not in file_name_dict:
            file_name_dict[file_name] = []
        file_name_dict[file_name].append(uid)

    for file_name in file_name_dict:
        logger.info(f"Loading {file_name}")
        filter_obj = Filters({"uidlist": WhitelistFilter(file_name_dict[file_name])})
        file_path = find_binary_file(data_dir, "**/*.pgdf", file_name)
        if file_path is None:
            report.add_error(FileNotFoundError(f"File {file_name} not found in {data_dir}."))
            continue
        file_data = load_pamguard_binary_file(file_path, filters=filter_obj, report = report)
        file_data.add_file_info()
        if len(file_data.data) != len(file_name_dict[file_name]):
            report.add_error(MultiFileException(file_name, f"Expected {len(file_name_dict[file_name])} items in {file_name}, found {len(file_data.data)}."))
        event_data.extend(file_data.data)

    return event_data, report


from pypamguard.chunks.standard import StandardModule
from pypamguard.core.readers import *

class DIFARProcessing(StandardModule):

    _minimum_version = 2 # as at 9 Jul 2025

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clip_start: np.int16 = None
        self.display_sample_rate: np.float32 = None
        self.demuxed_length: np.int32 = None
        self.amplitude: np.float32 = None
        self.gain: np.float32 = None
        self.sel_angle: np.float32 = None
        self.sel_freq: np.float32 = None
        self.species_code: str = None
        self.tracked_group: str = None
        self.max_val: np.float32 = None
        self.demux_data: np.ndarray = None
        self.num_matches: np.int16 = None
        self.latitude: np.float32 = None
        self.longitude: np.float32 = None
        self.errors: np.ndarray = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        self.clip_start = br.bin_read(DTYPES.INT64)
        self.display_sample_rate = br.bin_read(DTYPES.FLOAT32)
        self.demuxed_length = br.bin_read(DTYPES.INT32)
        self.amplitude = br.bin_read(DTYPES.FLOAT32)
        self.gain = br.bin_read(DTYPES.FLOAT32)
        self.sel_angle = br.bin_read(DTYPES.FLOAT32)
        self.sel_freq = br.bin_read(DTYPES.FLOAT32)
        self.species_code = br.string_read()
        self.tracked_group = br.string_read()
        self.max_val = br.bin_read(DTYPES.FLOAT32)
        if self.demuxed_length == 0: self.demux_data = 0
        else: self.demux_data = br.bin_read((DTYPES.INT16, lambda x: x * self.max_val / 32767), shape=(3, self.demuxed_length))
        self.num_matches = br.bin_read(DTYPES.INT16)
        if self.num_matches > 0:
            self.latitude = br.bin_read(DTYPES.FLOAT32)
            self.longitude = br.bin_read(DTYPES.FLOAT32)
            self.errors = br.bin_read(DTYPES.FLOAT32, shape=(2,))
            self.match_chan, self.match_time = br.bin_read([DTYPES.INT16, DTYPES.INT64], shape=(self.num_matches))
        else:
            self.latitude = 0
            self.longitude = 0
            self.errors = [0, 0, 0]



from pypamguard.chunks.base import BaseChunk
from pypamguard.core.readers import *
from pypamguard.chunks.generics import GenericChunkInfo
from pypamguard.utils.constants import IdentifierType
from pypamguard.core.exceptions import CriticalException, FileCorruptedException

class StandardChunkInfo(GenericChunkInfo):
    
    def __init__(self, file_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length: int = None
        self.identifier: int = None
        self.file_path: str = file_path

    def _process(self, br: BinaryReader):
        self.length, self.identifier = br.bin_read([DTYPES.INT32, DTYPES.INT32])
        if self.length < 0 or (self.identifier < 0 and self.identifier not in IdentifierType._value2member_map_):
            raise FileCorruptedException(br)



from abc import ABC, abstractmethod

from pypamguard.chunks.base import BaseChunk
from .genfileheader import GenericFileHeader

class GenericFileFooter(BaseChunk, ABC):
    
    def __init__(self, file_header: GenericFileHeader):
        super().__init__()
        self._file_header = file_header

        self.length: int = None
        self.identifier: int = None



from abc import ABC, abstractmethod

from pypamguard.chunks.base import BaseChunk

from .genfileheader import GenericFileHeader
from .genmoduleheader import GenericModuleHeader
from .genmodulefooter import GenericModuleFooter
from .genbackground import GenericBackground

from pypamguard.core.filters import FILTER_POSITION, Filters

class GenericModule(BaseChunk, ABC):

    _minimum_version = 0
    _maximum_version = None
    _header: GenericModuleHeader = None
    _footer: GenericModuleFooter = None # store the class of the footer
    _background: GenericBackground = None

    def __init__(self, file_header: GenericFileHeader, module_header: GenericModuleHeader, filters: Filters, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._file_header = file_header
        self._module_header = module_header
        self._filters = filters
  


from .baseannotation import BaseAnnotation
from pypamguard.chunks.annotations.submodules import ModelData
from pypamguard.core.readers import *

class DLAnnotation(BaseAnnotation):

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.n_models: np.int16 = None
        self.models: np.ndarray[ModelData] = None

    def _process(self, br: BinaryReader, *args, **kwargs):
        super()._process(br, *args, **kwargs)
        self.n_models = br.bin_read(DTYPES.INT16)
        self.models = np.ndarray(self.n_models, dtype=ModelData)
        for i in range(self.n_models):
            model = ModelData()
            model.process(br, *args, **kwargs)
            self.models[i] = model



from pypamguard.core.readers import *
from pypamguard.chunks.standard import StandardModule

class AISProcessing(StandardModule):

    _minimum_version = 1 # As at 9 Jul 2025

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mmsi_number: np.int32 = None
        self.fill_bits: np.int16 = None
        self.char_data: str = None
        self.ais_channel: str = None
    
    def _process(self, br, chunk_info):
        super()._process(br, chunk_info)
        self.mmsi_number = br.bin_read(DTYPES.INT32)
        self.fill_bits = br.bin_read(DTYPES.INT16)
        self.char_data = br.string_read()
        self.ais_channel = br.string_read()



from pypamguard.core.readers import *
from pypamguard.chunks.standard import StandardModule, StandardModuleHeader
from pypamguard.chunks.generics import GenericChunkInfo

class LongTermSpectralAverageHeader(StandardModuleHeader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def _process(self, br: BinaryReader, chunk_info: GenericChunkInfo):
        super()._process(br, chunk_info)

        if self.binary_length != 0:
            self.fft_length = br.bin_read(DTYPES.INT32)
            self.fft_hop = br.bin_read(DTYPES.INT32)
            self.interval_seconds = br.bin_read(DTYPES.INT32)

class LongTermSpectralAverage(StandardModule):

    _header = LongTermSpectralAverageHeader
    _minimum_version = 2

    a = 127 * 2 / np.log(32767)
    b = -127

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.end_millis: int = None
        self.end_date: datetime = None
        self.n_fft: np.int32 = None
        self.max_val: np.float32 = None
        self.bytes_data: np.ndarray = None
        self.data: np.ndarray = None


    def _process(self, br: BinaryReader, chunk_info: GenericChunkInfo):
        super()._process(br, chunk_info)
        
        self.end_date, self.end_millis = br.timestamp_read()

        self.n_fft = br.bin_read(DTYPES.INT32)
        self.max_val = br.bin_read(DTYPES.FLOAT32)

        # The type of the data is int8 but it needs to be promoted to int16
        # to support the arithmetic in the line after.
        self.data = br.bin_read((DTYPES.INT8, lambda x: np.exp((x.astype(np.int16,copy=False) - self.b) / self.a) * self.max_val / 32767), shape=(int(self._module_header.fft_length/2),))



import enum
from contextlib import contextmanager

class Verbosity(enum.IntEnum):
    CRITICAL = 0
    ERROR = 1
    WARNING = 2
    INFO = 3
    DEBUG = 4

@contextmanager
def logger_config(verbosity: Verbosity):
    old_verbosity = logger.verbosity
    logger.set_verbosity(verbosity)
    yield
    logger.set_verbosity(old_verbosity)


class Logger:

    def __init__(self):
        self.verbosity = Verbosity.INFO

    def set_verbosity(self, verbosity):
        self.verbosity = verbosity

    def _log(self, level, message):
        lines = str(message).split('\n')
        print(f"[{level}] {lines[0]}")
        for line in lines[1:]:
            print(f"\t{line}")

    def info(self, message):
        if self.verbosity >= Verbosity.INFO:
            self._log("INFO", message)

    def debug(self, message, br = None):
        if self.verbosity >= Verbosity.DEBUG:
            if br: message = "{0} bytes: {1}".format(br.tell(), message)
            self._log("DEBUG", message)

    def warning(self, message):
        if self.verbosity >= Verbosity.WARNING:
            self._log("WARNING", message)

    def error(self, message):
        self._log("ERROR", message)

logger = Logger()
